from osw.auth import CredentialManager
from osw.core import OSW, model
from osw.wtsite import WtSite

# run with: tox -e test -- --wiki_domain domain --wiki_username user --wiki_password pass


def test_file_controller(wiki_domain, wiki_username, wiki_password):
    cm = CredentialManager()
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=wiki_domain, username=wiki_username, password=wiki_password
        )
    )
    wtsite = WtSite(WtSite.WtSiteConfig(iri=wiki_domain, cred_mngr=cm))
    # site = mwclient.Site(host=wiki_domain, scheme="http")
    # site.login(username=wiki_username, password=wiki_password)
    # wtsite = WtSite(WtSite.WtSiteLegacyConfig(site=site))
    osw_obj = OSW(site=wtsite)

    list_of_categories = [
        "Category:OSW11a53cdfbdc24524bf8ac435cbf65d9d",  # WikiFile
        "Category:OSW3e3f5dd4f71842fbb8f270e511af8031",  # LocalFile
    ]
    for i, cat in enumerate(list_of_categories):
        mode = "append"
        if i == 0:
            mode = "replace"
        osw_obj.fetch_schema(OSW.FetchSchemaParam(schema_title=cat, mode=mode))

    run_test(osw_obj, "tests/integration/test.svg", "tests/integration/test2.svg")
    run_test(osw_obj, "tests/integration/test.png", "tests/integration/test2.png")


def run_test(osw_obj, file_path_1, file_path_2):
    from osw.controller.file.local import LocalFileController
    from osw.controller.file.wiki import WikiFileController

    lf = model.LocalFile(label=[model.Label(text="Test File")]).cast(
        LocalFileController, path=file_path_1
    )
    # Does not work due to missing attributes 'title' and 'osw':
    # wf = WikiFileController.from_local(lf)
    wf = WikiFileController(label=[model.Label(text="Test File")], osw=osw_obj)
    wf.put_from(lf)
    # print(wf)

    title = wf.namespace + ":" + wf.title
    # title = "File:OSW1b3fea404fe344c78ffd2d7a46bb468e.svg"
    wf2 = osw_obj.load_entity(title).cast(WikiFileController, osw=osw_obj)
    lf2 = LocalFileController(path=file_path_2)
    lf2.put_from(wf2)
    # wf2.get_to(lf2)

    lf2.delete()
    wf2.delete()



from __future__ import annotations

import getpass
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Union
from warnings import warn

import yaml
from pydantic.v1 import PrivateAttr

from osw.defaults import paths as default_paths
from osw.model.static import OswBaseModel

if TYPE_CHECKING:
    PossibleFilePath = Path
else:
    from osw.custom_types import PossibleFilePath


class CredentialManager(OswBaseModel):
    """Handles credentials"""

    cred_filepath: Optional[
        Union[Union[str, PossibleFilePath], List[Union[str, PossibleFilePath]]]
    ]
    """Filepath to yaml file with credentials for osw and connected services"""
    cert_filepath: Optional[
        Union[Union[str, PossibleFilePath], List[Union[str, PossibleFilePath]]]
    ]
    """Filepath to the certificates for osw and connected services"""

    _credentials: List[BaseCredential] = PrivateAttr([])
    """in memory credential store"""

    class BaseCredential(OswBaseModel):
        """Abstract base class for credentials"""

        iri: str
        """the iri the credential is valid for"""

    class UserPwdCredential(BaseCredential):
        """a username - password credential"""

        username: str
        """the user identifier"""
        password: str
        """the users password"""

    class OAuth1Credential(BaseCredential):
        """OAuth1 credentials. See
        https://requests-oauthlib.readthedocs.io/en/latest/oauth1_workflow.html"""

        consumer_token: str
        """consumer token"""
        consumer_secret: str
        """consumer secret """
        access_token: str
        """access token"""
        access_secret: str
        """access secret"""

    class CredentialFallback(Enum):
        """Modes of handling missing credentials

        Attributes
        ----------
        none:
            throw error
        ask:
            use getpass to ask for credentials
        """

        ask = "ask"  # use getpass to ask for credentials
        none = "none"  # throw error

    class CredentialConfig(OswBaseModel):
        """Reads credentials from a yaml file"""

        iri: str
        """internationalized resource identifier / address of the service, may contain
        protocol, domain, port and path matches by "contains" returning the shortest
        match"""
        fallback: Optional[CredentialManager.CredentialFallback] = "none"
        """The fallback strategy if no credential was found for the given origin"""

    def __init__(self, **data):
        super().__init__(**data)
        if self.cred_filepath:
            if not isinstance(self.cred_filepath, list):
                self.cred_filepath = [self.cred_filepath]
            self.cred_filepath = [Path(fp) for fp in self.cred_filepath if fp != ""]
        # Make sure to at least warn the user if they pass cred_filepath instead of
        # cred_filepath
        attribute_names = self.__dict__.keys()
        unexpected_kwargs = [key for key in data.keys() if key not in attribute_names]
        if unexpected_kwargs:
            warn(f"Unexpected keyword argument(s): {', '.join(unexpected_kwargs)}")

    def get_credential(self, config: CredentialConfig) -> BaseCredential:
        """Reads credentials from a yaml file or the in memory store

        Parameters
        ----------
        config:
            see CredentialConfig

        Returns
        -------
        credential :
            Credential, contain attributes 'username' and 'password' and
            the matching iri.
        """

        _file_credentials: List[CredentialManager.BaseCredential] = []
        if self.cred_filepath:
            filepaths = self.cred_filepath
            if type(filepaths) is not list:
                filepaths = [filepaths]

            for filepath in filepaths:
                if not filepath.exists():
                    continue
                with open(filepath, "r", encoding="utf-8") as stream:
                    try:
                        accounts = yaml.safe_load(stream)
                        if accounts is None:  # Catch empty file
                            continue
                        for iri in accounts.keys():
                            if (
                                "username" in accounts[iri]
                                and "password" in accounts[iri]
                            ):
                                cred = CredentialManager.UserPwdCredential(
                                    username=accounts[iri]["username"],
                                    password=accounts[iri]["password"],
                                    iri=iri,
                                )
                                _file_credentials.append(cred)
                            if (
                                "consumer_token" in accounts[iri]
                                and "consumer_secret" in accounts[iri]
                                and "access_token" in accounts[iri]
                                and "access_secret" in accounts[iri]
                            ):
                                cred = CredentialManager.OAuth1Credential(
                                    consumer_token=accounts[iri]["consumer_token"],
                                    consumer_secret=accounts[iri]["consumer_secret"],
                                    access_token=accounts[iri]["access_token"],
                                    access_secret=accounts[iri]["access_secret"],
                                    iri=iri,
                                )
                                _file_credentials.append(cred)
                    except yaml.YAMLError as exc:
                        print(exc)

        match_iri = ""
        cred = None
        creds = _file_credentials + self._credentials
        for _cred in creds:
            iri = _cred.iri
            if config.iri in iri:
                if match_iri == "" or len(match_iri) > len(
                    iri
                ):  # use the less specific match
                    match_iri = iri
                    cred = _cred

        if cred is None:
            if config.fallback is CredentialManager.CredentialFallback.ask:
                print(
                    f"No credentials for {config.iri} found. "
                    f"Please use the prompt to login"
                )
                username = input("Enter username: ")
                password = getpass.getpass("Enter password: ")
                cred = CredentialManager.UserPwdCredential(
                    username=username, password=password, iri=config.iri
                )
                self.add_credential(cred)
                if self.cred_filepath:
                    self.save_credentials_to_file()
        return cred

    def add_credential(self, cred: BaseCredential):
        """adds a credential to the in memory store

        Parameters
        ----------
        cred
            the credential to add
        """
        self._credentials.append(cred)

    def iri_in_credentials(self, iri: str) -> bool:
        """checks if a credential for a given iri exists

        Parameters
        ----------
        iri
            the iri to check

        Returns
        -------
        bool
            True if a credential exists for the given iri
        """
        for cred in self._credentials:
            if cred.iri == iri:
                return True
        return False

    def iri_in_file(self, iri: str) -> bool:
        """checks if a credential for a given iri exists in the file

        Parameters
        ----------
        iri
            the iri to check

        Returns
        -------
        bool
            True if a credential exists for the given iri
        """
        if self.cred_filepath:
            for fp in self.cred_filepath:
                if fp != "":
                    if Path(fp).exists():
                        with open(fp, "r", encoding="utf-8") as stream:
                            try:
                                accounts = yaml.safe_load(stream)
                                if accounts is None:  # Catch empty file
                                    continue
                                for iri_ in accounts.keys():
                                    if iri_ == iri:
                                        return True
                            except yaml.YAMLError as exc:
                                print(exc)
        return False

    def save_credentials_to_file(
        self,
        filepath: Union[str, PossibleFilePath] = None,
        set_cred_filepath: bool = False,
    ):
        """Saves the in memory credentials to a file

        Parameters
        ----------
        filepath
            The filepath to save the credentials to. If None, the filepath specified
            in the CredentialManager is used.  If cred_filepath and filepath are None,
            the default path is used. If the file does not exist, it is created.
        set_cred_filepath
            If True, the cred_filepath is set to the given filepath. If False, the
            cred_filepath of the CredentialManager is not changed.
        """
        cred_filepaths = [filepath]
        """The filepath to save the credentials to."""
        if filepath is None:
            cred_filepaths = self.cred_filepath
            if self.cred_filepath is None:
                cred_filepaths = [default_paths.cred_filepath]
        if set_cred_filepath:
            # Creates error if file does not exist -> Using custom FilePath
            self.cred_filepath = cred_filepaths
        for fp in cred_filepaths:
            file = Path(fp)
            if not file.parent.exists():
                file.parent.mkdir(parents=True)
            data = {}
            file_already_exists = file.exists()
            if file_already_exists:
                data = yaml.safe_load(file.read_text(encoding="utf-8"))
                if data is None:  # Catch empty file
                    data = {}
            for cred in self._credentials:
                data[cred.iri] = cred.dict(exclude={"iri"})
            with open(fp, "w", encoding="utf-8") as stream:
                yaml.dump(data, stream)
            if file_already_exists:
                print(f"Credentials file updated at '{fp.resolve()}'.")
            else:
                print(f"Credentials file created at '{fp.resolve()}'.")

        # Creating or updating .gitignore file in the working directory
        cwd = Path.cwd()
        potential_fp = [
            cwd / ".gitignore",
            cwd.parent / ".gitignore",
        ]
        gitignore_fp = potential_fp[0]
        # Stops if a .gitignore file is found
        for fp in potential_fp:
            if fp.exists():
                gitignore_fp = fp
                break
        # Creates a .gitignore file if none is found
        if not gitignore_fp.exists():
            if not gitignore_fp.parent.exists():
                gitignore_fp.parent.mkdir(parents=True)
            gitignore_fp.touch()
        # Reads the .gitignore file
        with open(gitignore_fp, "r") as stream:
            content = stream.read()
        comment_set = False
        osw_dir_added = False
        # For every file path in the list of credentials file paths
        for _ii, fp in enumerate(cred_filepaths):
            to_add = ""
            if default_paths.osw_files_dir in fp.parents and not osw_dir_added:
                print(
                    f"Adding '{default_paths.osw_files_dir}' to gitignore file "
                    f"'{gitignore_fp}'."
                )
                containing_gitignore = gitignore_fp.parent.absolute()

                if containing_gitignore in default_paths.osw_files_dir.parents:
                    # If the default_path.osw_files_dir is a subdirectory of the directory
                    # containing the .gitignore file, add the relative path to the
                    # .gitignore file
                    rel = default_paths.osw_files_dir.relative_to(containing_gitignore)
                    to_add = f"\n*/{str(rel.as_posix())}/*"
                else:
                    # Test if the default_path.osw_files_dir is a subdirectory of the
                    # directory containing the .gitignore file
                    to_add = (
                        f"\n*/{default_paths.osw_files_dir.absolute().as_posix()}/*"
                    )
                osw_dir_added = True
            else:
                print(f"Adding '{fp.name}' to gitignore file '{gitignore_fp}'.")
                to_add = f"\n*/{fp.name}"
            if not to_add or to_add in content:
                continue

            with open(gitignore_fp, "a") as stream:
                # Only add comment if not already set
                comment = (
                    "\n# Automatically added by osw.auth.CredentialManager."
                    "save_credentials_to_file:"
                )
                if not comment_set and comment not in content:
                    stream.write(comment)
                    comment_set = True
                stream.write(to_add)


CredentialManager.CredentialConfig.update_forward_refs()



import uuid

import osw.model.entity as model
from osw.utils.regex import count_match_groups
from osw.utils.strings import pascal_case
from osw.utils.wiki import (
    get_full_title,
    get_namespace,
    get_osw_id,
    get_title,
    get_uuid,
    is_empty,
    namespace_from_full_title,
    title_from_full_title,
)


# osw.utils.wiki
def test_get_osw_id():
    uuid_ = uuid.uuid4()
    osw_id = f"OSW{str(uuid_).replace('-', '')}"
    assert get_osw_id(uuid_) == osw_id


def test_get_uuid():
    uuid_ = uuid.uuid4()
    osw_id = f"OSW{str(uuid_).replace('-', '')}"
    assert get_uuid(osw_id) == uuid_


def test_get_entity_namespace():
    class DummyClass(model.Entity):
        pass

    assert get_namespace(DummyClass) == "Category"

    e = model.Entity(label=[model.Label(text="MyItem")])
    assert get_namespace(e) is None

    # e = model.Item(label=[model.Label(text="MyItem")])
    # assert get_namespace(e) == "Item"

    # not yet loaded on default
    # e = model.Category(label=[model.Label(text="MyItem")])
    # assert get_namespace(e) == "Category"

    # e = model.Property(label=[model.Label(text="MyItem")])
    # assert get_namespace(e) == "Property"

    # e = model.WikiFile(label=[model.Label(text="MyItem")])
    # assert get_namespace(e) == "File"


def test_get_title():
    title = "MyItem"
    namespace = "Item"
    entity = model.Item(label=[model.Label(text=title)])
    second_entity = model.Item(label=[model.Label(text=title)])
    second_entity.meta = model.Meta(
        wiki_page=model.WikiPage(title=title, namespace=namespace)
    )
    assert get_title(entity) == get_osw_id(entity.uuid)
    assert get_title(second_entity) == title


def test_get_full_title():
    title = "MyItem"
    namespace = "Item"
    entity = model.Item(label=[model.Label(text=title)])
    second_entity = model.Item(label=[model.Label(text=title)])
    second_entity.meta = model.Meta(
        wiki_page=model.WikiPage(title=title, namespace=namespace)
    )
    second_entity.meta.wiki_page.namespace = namespace
    assert get_full_title(entity) == f"Item:{get_osw_id(entity.uuid)}"
    assert get_full_title(second_entity) == f"{namespace}:{title}"


def test_namespace_from_full_title():
    full_title = "Item:MyTitle"
    assert namespace_from_full_title(full_title) == "Item"


def test_title_from_full_title():
    full_title = "Item:MyTitle"
    assert title_from_full_title(full_title) == "MyTitle"


def test_is_empty():
    assert is_empty(None)
    assert is_empty("")
    assert is_empty([])
    assert is_empty({})
    assert not is_empty("String")
    assert not is_empty(["List"])
    assert not is_empty({"Dict": "Value"})


# osw.utils.strings
def test_count_match_groups():
    pat = r"(\\d+)(\\w+)\\((\\d+)\\)"
    assert count_match_groups(pat) == 3
    pat2 = r"((\\d+)(\\w+))"
    assert count_match_groups(pat2) == 3
    pat3 = r"(File:(OSW[0-9a-fA-F]{32})((?:\.[\w-]+)*))"
    assert count_match_groups(pat3) == 3


def test_strings_pascalcase():
    assert pascal_case("my string") == "MyString"
    assert pascal_case("my_string") == "MyString"
    assert pascal_case("MyString") == "MyString"


# osw.utils.regex_pattern
def test_regex_pattern():
    """This tests the regex pattern library. For all patterns in the library, the
    test_pattern method is called. If the method does not raise an exception, the
    pattern is considered correct. If no example_str and no expected_groups are
    provided, the test_pattern method raise a type error originating from the re
    module."""
    from osw.utils.regex_pattern import REGEX_PATTERN_LIB

    for _key, pattern in REGEX_PATTERN_LIB.items():
        assert pattern.test_pattern()



import os
import shutil
from io import TextIOBase
from pathlib import Path
from typing import IO, Any, Dict, List, Optional

from osw.controller.file.base import FileController
from osw.core import model


class LocalFileController(FileController, model.LocalFile):
    """File controller for local files"""

    label: Optional[List[model.Label]] = [model.Label(text="Unnamed file")]
    """The label of the file, defaults to 'Unnamed file'"""
    path: Path
    """The path to the file"""

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def from_other(
        cls, other: "FileController", path: Path, **kwargs
    ) -> "LocalFileController":
        return other.cast(cls, path=path, **kwargs)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._set_metadata()

    def get(self) -> IO:
        self._set_metadata()
        return open(self.path, "rb")

    def put(self, file: IO, **kwargs: Dict[str, Any]):
        self._set_metadata()
        mode = "wb"
        if isinstance(file, TextIOBase):
            mode = "w"
        with open(self.path, mode) as f:
            # f.write(file.read())  # In-memory - limited by available RAM
            shutil.copyfileobj(file, f)  # Chunked copy

    def delete(self):
        if os.path.exists(self.path):
            os.remove(self.path)

    def _set_metadata(self):
        file = LocalFileController.extract_metadata(self.path)
        self.label = file.label

    @classmethod
    def extract_metadata(cls, path: Path) -> model.LocalFile:
        file = model.LocalFile(label=[model.Label(text=path.name)])
        if os.path.exists(path):
            pass
            # stats = os.stat(path)
            # TODO: extract metadata
            # ... file.size = stats.st_size
        return file



import getpass
from typing import Dict, List, Optional, Tuple, Union

import mwclient
import yaml
from pydantic.v1 import FilePath

from osw.model.static import OswBaseModel
from osw.utils.util import parallelize

# try import functions from wikitext.py (relies on the extra dependency osw[wikitext])
try:
    from osw.utils.wikitext import create_flat_content_structure_from_wikitext  # noqa
    from osw.utils.wikitext import extend_page_list_with_recursive_dependencies  # noqa
    from osw.utils.wikitext import find_dependencies  # noqa
    from osw.utils.wikitext import find_dependencies_recursively  # noqa
    from osw.utils.wikitext import get_wikitext_from_flat_content_dict  # noqa
    from osw.utils.wikitext import get_wikitext_from_flat_content_structure  # noqa
    from osw.utils.wikitext import merge_wiki_page_text  # noqa;
    from osw.utils.wikitext import schemaJson2WikiJson  # noqa
    from osw.utils.wikitext import update_template_within_wikitext  # noqa
    from osw.utils.wikitext import wikiJson2SchemaJson  # noqa
    from osw.utils.wikitext import wikiJson2SchemaJsonRecursion  # noqa

except ImportError:
    print(
        "Hint: The extra dependency 'osw[wikitext]' "
        "is required for the full functionality of wiki_tools module."
    )


def read_domains_from_credentials_file(
    cred_filepath: Union[str, FilePath]
) -> Tuple[List[str], Dict[str, Dict[str, str]]]:
    """Reads domains and credentials from a yaml file

    Parameters
    ----------
    cred_filepath
        Path to the yaml file with the credentials

    Returns
    -------

    """
    with open(cred_filepath, "r", encoding="utf-8") as stream_:
        try:
            accounts_dict = yaml.safe_load(stream_)
            domains_list = list(accounts_dict.keys())
            if len(domains_list) == 0:
                raise ValueError("No domain found in accounts.pwd.yaml!")
            return domains_list, accounts_dict
        except yaml.YAMLError as exc_:
            print(exc_)


def read_credentials_from_yaml(
    password_file: Union[str, FilePath], domain: str = None
) -> dict:
    """Reads credentials from a yaml file

    Parameters
    ----------
    password_file :
        Path to the yaml file with the credentials.
    domain:
        Domain of the OSW instance, as specifed in the yaml file.

    Returns
    -------
    credentials :
        Dictionary with the credentials, expected to
        contain keys 'username' and 'password'.
    """
    if password_file != "":
        with open(password_file, "r") as stream:
            try:
                accounts = yaml.safe_load(stream)
                if domain is not None and domain in accounts.keys():
                    domain = domain
                elif len(accounts.keys()) > 0:
                    domain = list(accounts.keys())[0]
                    if len(accounts.keys()) > 0:
                        domain = list(accounts.keys())[0]
                user = accounts[domain]["username"]
                password = accounts[domain]["password"]
            except yaml.YAMLError as exc:
                print(exc)
    else:
        user = input("Enter bot username (username@botname)")
        password = getpass.getpass("Enter bot password")
    return {"username": user, "password": password}


def create_site_object(
    domain: str, password_file: Union[str, FilePath] = "", credentials: dict = None
) -> mwclient.client.Site:
    """
    Parameters
    ----------
    domain :
        Domain of the OSW instance, as specifed in the yaml file
    password_file :
        path to file with <username>\n<password>
    credentials :
        Dictionary with the credentials (username, password)
    Returns
    -------
    site : mwclient.client.Site
        Site object from mwclient lib
    """
    domain_dict = {
        "wiki-dev": {"Address": "wiki-dev.open-semantic-lab.org"},
        "onterface": {"Address": "onterface.open-semantic-lab.org:"},
    }
    if domain in domain_dict.keys():
        domain = domain_dict[domain]["Address"]

    site = mwclient.Site(domain, path="/w/")
    if credentials is None:
        credentials = read_credentials_from_yaml(password_file, domain)
    # else:
    #     credentials = credentials
    # Login with dictionary unpacking:
    # site.login(**credentials)
    # Explicit login:
    site.login(username=credentials["username"], password=credentials["password"])
    del credentials
    return site


class SearchParam(OswBaseModel):
    """Search parameters for semantic and prefix search"""

    query: Union[str, List[str]]
    parallel: Optional[bool] = None  # is set to true if query is a list longer than 5
    debug: Optional[bool] = True
    limit: Optional[int] = 1000

    def __init__(self, **data):
        super().__init__(**data)
        if not isinstance(self.query, list):
            self.query = [self.query]
        if len(self.query) > 5 and self.parallel is None:
            self.parallel = True
        if self.parallel is None:
            self.parallel = False


def prefix_search(
    site: mwclient.client.Site, text: Union[str, SearchParam]
) -> List[str]:
    """Standard query. Equivalent to the following mediawiki API call
    api.php?action=query&list=prefixsearch&pssearch=Star Wars.

    See https://www.mediawiki.org/wiki/API:Prefixsearch for details.

    Parameters
    ----------
    site :
        Site object from mwclient lib
    text :
        Query text or instance of SearchParam

    Returns
    -------
    page_list :
        List of page titles
    """
    if not isinstance(text, SearchParam):
        query = SearchParam(query=text)
    else:
        query = text

    def prefix_search_(single_text):
        page_list = list()
        result = site.api(
            "query",
            list="prefixsearch",
            pssearch=single_text,
            pslimit=query.limit,
            format="json",
        )
        if len(result["query"]["prefixsearch"]) == 0:
            if query.debug:
                print("No results")
        else:
            for page in result["query"]["prefixsearch"]:
                title = page["title"]
                if query.debug:
                    print(title)
                page_list.append(title)
        return page_list

    if query.parallel:
        query_results = parallelize(
            func=prefix_search_, iterable=query.query, flush_at_end=query.debug
        )
    else:
        query_results = [prefix_search_(single_text=sq) for sq in query.query]

    return [item for sublist in query_results for item in sublist]
    # todo: @Simon: a list of lists of strings (sublist for each query in query list)
    #  or a list of strings (results of all queries combined)?
    #  The last option would not change the behavior of the function, but would
    # return page_list  # original return


def semantic_search(
    site: mwclient.client.Site, query: Union[str, List[str], SearchParam]
) -> List[str]:
    """Semantic query

    Parameters
    ----------
    site :
        Site object from mwclient lib
    query :
        (List of) query text(s) or instance of SearchParam

    Returns
    -------
    page_list:
        List of page titles
    """
    if not isinstance(query, SearchParam):
        query = SearchParam(query=query)

    def semantic_search_(single_query):
        page_list = list()
        single_query += f"|limit={query.limit}"
        result = site.api("ask", query=single_query, format="json")
        if len(result["query"]["results"]) == 0:
            if query.debug:
                print("Query '{}' returned no results".format(single_query))
        else:
            if query.debug:
                print(
                    "Query '{}' returned {} results".format(
                        single_query, len(result["query"]["results"])
                    )
                )
            for page in result["query"]["results"].values():
                # why do we do the following?
                if "printouts" in page:
                    title = page["fulltext"]
                    if "#" not in title and query.debug:
                        print(title)
                        # original position of "page_list.append(title)" line
                    page_list.append(title)
        return page_list

    if query.parallel:
        query_results = parallelize(
            func=semantic_search_, iterable=query.query, flush_at_end=query.debug
        )
    else:
        query_results = [semantic_search_(single_query=sq) for sq in query.query]

    return [item for sublist in query_results for item in sublist]
    # todo: @Simon: a list of lists of strings (sublist for each query in query list)
    #  or a list of strings (results of all queries combined)?
    #  The last option would not change the behavior of the function, but would
    # return page_list  # original return


def search_wiki_page(title: str, site: mwclient.client.Site):
    """Page search wrapper that adds exact match functionality with ignore-case on
    top of the  prefix_search()'s functionality.

    Parameters
    ----------
    title :
        Title of the wiki page, e.g., User:Someone1234
    site :
        Site object from mwclient lib

    Returns
    -------
    result_dict : dict
    """
    result = site.api(
        "query", list="prefixsearch", pssearch=title, pslimit=1000, format="json"
    )
    if len(result["query"]["prefixsearch"]) == 0:
        return {"Result": False, "List": list()}
    else:
        exact_match = False
        page_title_list = list()
        for page in result["query"]["prefixsearch"]:
            page_title = page["title"]
            page_title_list.append(page_title)
            if page_title.lower() == title.lower():
                exact_match = True
        result_dict = {
            "Result": True,
            "List": page_title_list,
            "Exact match": exact_match,
        }
        return result_dict


def create_or_update_wiki_page_with_template(
    title, content, site, overwrite_with_empty=False
):
    """Creates a wiki page with a template included in the content. If the page does
    already exist, the parameters within the template are update

    Parameters
    ----------
    title : str
        Title of the wiki page, e.g., User:Someone1234
    content : str
    site : mwclient.client.Site
        Site object from mwclient lib
    overwrite_with_empty : bool
        Decided whether a template parameter's value in an preexisting page is
        overwritten with an empty value

    Returns
    -------
    success : bool
    """
    search_result = search_wiki_page(title, site)
    if search_result["Result"]:
        existing_page = site.pages[title]
        existing_text = existing_page.text()
        # update the page's content (template only)
        updated_content = update_template_within_wikitext(
            text=existing_text,
            template_text=content,
            overwrite_with_empty=overwrite_with_empty,
        )
        success = create_or_overwrite_wiki_page(title, updated_content, site)
    else:
        # just create the page
        success = create_or_overwrite_wiki_page(title, content, site)
    return success


def edit_wiki_page_with_content_merge(title, new_content, site, template_name):
    """Edits an existing wiki page, while merging the passed new content with the
    content of the existing page

    Parameters
    ----------
    title : str
        Title of the wiki page, e.g., User:Someone1234
    new_content : str
    site : mwclient.client.Site
        Site object from mwclient lib
    template_name : str

    Returns
    -------
    success : bool
    """
    search_result = search_wiki_page(title, site)
    if search_result["Result"]:
        source_page = site.pages[title]
        source_page_content = source_page.text()
        # todo: test function
        new_content = merge_wiki_page_text(
            new_content, source_page_content, template_name=template_name
        )
        target_page = site.pages[title]
        target_page.edit(new_content, "[bot] update of page content")
        success = True
    else:
        success = False
    return success


def get_file_info_and_usage(
    site: mwclient.client.Site, title: Union[str, List[str], SearchParam]
) -> List[Dict[str, Union[Dict[str, str], List[str]]]]:
    """(For 'File' pages only) Get information about the file and its usage

    Parameters
    ----------
    site:
        Site object from mwclient lib.
    title:
        Title(s) of the wiki page(s) or instance of SearchParam.

    Returns
    -------
    result:
        Dictionary with page titles as keys and nested dictionary with keys 'info' and
        'usage'.

    Notes
    -----
    Query to reproduce:
        action=query
        format=json
        prop=imageinfo|fileusage
        titles=File%3AOSW857d85031d85425aa94db8b4720e84b7.png
        &iiprop=timestamp%7Cuser&fulimit=5000"

    Resources
    ---------
    Use the sandbox to design and test the queries:
    https://demo.open-semantic-lab.org/wiki/Special:ApiSandbox
    """
    if not isinstance(title, SearchParam):
        query = SearchParam(query=title, debug=False)
    else:  # SearchParam
        query = title

    def get_file_info_and_usage_(single_title):
        api_request_result = site.api(
            action="query",
            format="json",
            prop="imageinfo|fileusage",
            titles=single_title,
            iiprop="timestamp|user",
            fulimit=query.limit,
        )
        using_pages = []
        file_info = {
            "title": single_title,
            "author": "File not found or no creation logged",
            "timestamp": "File not found or no creation logged",
            "editor": [],
            "editing_timestamp": [],
        }

        if len(api_request_result["query"]["pages"]) == 0:
            if query.debug:
                print(f"Page not found: '{single_title}'!")
        else:
            image_info: List[Dict[str, str]] = []
            file_usage: List[Dict[str, Union[str, int]]] = []
            for _page_id, page_dict in api_request_result["query"]["pages"].items():
                if page_dict["title"] == single_title:
                    image_info = page_dict.get("imageinfo", [])
                    file_usage = page_dict.get("fileusage", [])
            if len(image_info) != 0:
                file_info["author"] = image_info[0]["user"]
                file_info["timestamp"] = image_info[0]["timestamp"]
                for ii in image_info:
                    file_info["editor"].append(ii["user"])
                    file_info["editing_timestamp"].append(ii["timestamp"])
            if file_usage is not None:
                for fu_page_dict in file_usage:
                    using_pages.append(fu_page_dict["title"])
            if query.debug:
                # todo: find out why this message is printed (sometimes) when using the
                #  redirect,  which messes up the Progressbar
                #  printed messages do not appear in the MessageBuffer
                print(f"File info for '{single_title}' retrieved.")
        return {"info": file_info, "usage": using_pages}

    if query.parallel:
        api_request_results = parallelize(
            func=get_file_info_and_usage_,
            iterable=query.query,
            flush_at_end=query.debug,
        )
    else:
        api_request_results = [
            get_file_info_and_usage_(single_title=st) for st in query.query
        ]

    return api_request_results


def search_redirection_sources(
    site: mwclient.client.Site, target_title: str, debug: bool = False
):
    """Returns a list of pages redirecting to the page with target_title per #REDIRECT
    [[target]] syntax

    Parameters
    ----------
    site :
        Site object from mwclient lib
    target_title :
        Title of the target wiki page
    debug:
        Whether to print debugging messages

    Returns
    -------
    page_list : list of pages redirecting to the page with target_title
    """
    page_list = []
    result = site.api("query", titles=target_title, prop="redirects", format="json")
    if len(result["query"]["pages"]) == 0:
        if debug:
            print("No results")
    else:
        for page in result["query"]["pages"]:
            if "redirects" not in result["query"]["pages"][page]:
                if debug:
                    print("No results")
            else:
                for redirecting_source in result["query"]["pages"][page]["redirects"]:
                    title = redirecting_source["title"]
                    page_list.append(title)
    return page_list


def create_or_overwrite_wiki_page(title, content, site):
    """Creates a page with the passed title and content. If the page already exists,
    the prior content is replaced with the passed content.

    Parameters
    ----------
    title : str
        Title of the wiki page, e.g., User:Someone1234
    content : str
    site : mwclient.client.Site
        Site object from mwclient lib

    Returns
    -------
    success : bool
    """
    target_page = site.pages[title]
    target_page.edit(content, "[bot] create page")
    success = True
    return success


def delete_wiki_page(title, site, reason):
    """Deletes the wiki page with the passed title, if it was found (exact match!),
    otherwise returns False

    Parameters
    ----------
    title : str
        Title of the wiki page, e.g., User:Someone1234
    site : mwclient.client.Site
        Site object from mwclient lib
    reason : str

    Returns
    -------
    success : bool
    """
    try:
        page = site.pages[title]
        page.delete(reason=reason, watch=False, unwatch=True, oldimage=False)
        success = True
    except mwclient.errors.APIError:
        success = False
    return success


def copy_wiki_page(title0, title1, site0, site1, overwrite=True):
    """

    Parameters
    ----------
    title0 : str
        Title of the source page
    title1 : str
        Title of the target page
    site0 : mwclient.client.Site
        Source site object from mwclient lib
    site1 : mwclient.client.Site
        Target site object from mwclient lib
    overwrite : bool
        Whether to overwrite existing pages at target site

    Returns
    -------
    success: bool

    """
    if title0.lower() == title1.lower() and site0 == site1:
        # copy on it self = no action necessary
        success = True
    else:
        page0 = site0.pages[title0]
        content = page0.text()
        if overwrite:
            success = create_or_overwrite_wiki_page(title1, content, site1)
        else:
            search_result = search_wiki_page(title1, site1)
            if (
                search_result["Result"] and search_result["Exact match"]
            ):  # page already exists
                success = False
            else:
                # search_result["Result"] == True/False
                # search_result["Exact match"] == False
                success = create_or_overwrite_wiki_page(title1, content, site1)
    return success


def copy_list_of_wiki_pages(title_list, site0, site1, overwrite, callback=None):
    """

    Parameters
    ----------
    title_list : list
    site0 : mwclient.client.Site
        Source site object from mwclient lib
    site1 : mwclient.client.Site
        Target site object from mwclient lib
    overwrite : bool
        Whether to overwrite existing pages at target site
    callback : NoneType or function
        Function passed over, to perform operation on the titles of the source pages and
        to be passed as title of the
        target pages. See examples below.
        Example functions:
            capitalize = lambda x: x.capitalize()
            def change_namespace(title, namespace):
                if ":" in namespace:
                    namespace = namespace.split(":")[0]
                if ":" in title:
                    splits = title.split(":")
                    old_name = splits[1].capitalize()
                    new_title = namespace + ":" + old_name
                else:
                    new_title = namespace + ":" + title.capitalize()
                return new_title
        Examples of passing a function as parameter:
            callback = capitalize
            callback = lambda x: x.lower()

    Returns
    -------
    results_dict : dict
        Dictionary, containing the results of the copying operations
    """

    success_list = list()
    fail_list = list()
    for title0 in title_list:
        if callback is None:
            title1 = title0
        else:
            title1 = callback(title0)
        success = copy_wiki_page(title0, title1, site0, site1, overwrite)
        if success:
            success_list.append(title1)
        else:
            fail_list.append(title1)
    results_dict = {
        "Successfully copied pages": success_list,
        "Pages failed to copy": fail_list,
    }
    return results_dict



import json
import os
import re
import uuid
from typing import Dict, List, Literal, Optional, Type

from pydantic.v1 import PrivateAttr
from pyld import jsonld
from rdflib import Graph
from typing_extensions import deprecated

from osw.core import OSW, model
from osw.model.static import OswBaseModel
from osw.utils.strings import pascal_case
from osw.utils.wiki import get_namespace
from osw.wtsite import WtSite


class ParserSettings(OswBaseModel):
    """Settings for the ontology parser"""

    flatten_subclassof_restrictions: Optional[bool] = False
    """If True, the parser will directly asign the property and the value class of the
    restriction to the restricted class
    """
    subclassof_restrictions_keyword: Optional[str] = "restrictions"
    """Keyword to separate the owl restrictions from rdfs subclassof. Only used if
    flatten_subclassof_restrictions is False."""
    ensure_multilang: Optional[List[str]] = [
        "label",
        "prefLabel",
        "altLabel",
        "comment",
        "description",
    ]
    """List of properties that should be multilang / have a @lang containter. If the
    property is not multilang, the parser will create a multilang property with "en" as
    default language."""
    ensure_array: Optional[List[str]] = [
        "rdf_type",
        "label",
        "prefLabel",
        "altLabel",
        "comment",
        "description",
        "subclass_of",
        "restrictions",
        "range",
    ]
    """List of properties that should be arrays. If the property is not an array, the
    parser will create an array with the value as single element."""
    map_uuid_iri: Optional[List[str]] = [
        "subclass_of",
        "on_property",
        "some_values_from",
        "all_values_from",
        "domain",
        "range",
        "subproperty_of",
    ]
    """List of properties whose values should be mapped to (OSW) UUIDs. If the value is
    not a UUID, the parser will create a UUIDv5 from the value with the URL
    namespace."""
    remove_unnamed: Optional[List[str]] = []  # ['subclass_of'] #, 'equivalentClass']
    """List of properties whose values should be removed if they are unnamed nodes
    ("_:...")"""


class ImportConfig(OswBaseModel):
    """Configuration for the ontology importer"""

    file: str
    """the path or url to the ontology file"""
    file_format: Literal["xml", "n3", "nt", "trix"] = "n3"
    """the serialization format. for turtle, use 'n3' (see rdflib docs)"""
    ontology_name: str
    """the name of the ontology"""
    ontologies: List[model.OwlOntology]
    """Ontology metadata, including imported ontologies"""
    import_mapping: Optional[Dict[str, str]] = {}
    """Mapping of imported ontologies iri to their resolveable file url in case both
    are not identical"""
    base_class: Type[OswBaseModel]
    """Base class for the ontology model. For OWL Ontologies, this should be
    model.OwlClass or a subclass of it."""
    base_class_title: Optional[str] = (
        "Category:OSW725a3cf5458f4daea86615fcbd0029f8"  # OwlClass
    )
    """Title of the base class schema. Defaults to OwlClass"""
    meta_class_title: Optional[str] = (
        "Category:OSW379d5a1589c74c82bc0de47938264d00"  # OwlThing
    )
    """Title of the meta class schema. Defaults to OwlThing"""
    dump_files: Optional[bool] = False
    """If True, the parsed ontology will be dumped to a jsonld file"""
    dump_path: Optional[str] = None
    """Path to the directory where the parsed ontology will be dumped"""
    dry_run: Optional[bool] = False
    """If True, the parsed ontology will not be imported into the wiki"""
    property_naming_policy: Optional[Literal["UUID", "label", "prefixed_label"]] = (
        "prefixed_label"
    )
    """Policy for naming properties. If "UUID", the property will be named with a
    UUIDv5.
    If "label", the property will be named with the label of the property.
    If "prefixed_label", the property will be named with the label of the property
    prefixed with the label of the ontology."""
    property_naming_prefix_delimiter: Optional[str] = ":"
    """Delimiter for the prefixed_label property naming policy"""

    class Config:
        arbitrary_types_allowed = True


class OntologyImporter(OswBaseModel):
    """Class for importing ontologies into OSW"""

    osw: OSW
    """A OSW instance"""
    parser_settings: Optional[ParserSettings] = ParserSettings()
    """Settings for the ontology parser"""
    import_config: Optional[ImportConfig] = None
    """Configuration for the ontology importer"""

    _context: Optional[Dict] = PrivateAttr()
    _id_dict: Optional[Dict] = PrivateAttr()
    _iri_dict: Optional[Dict] = PrivateAttr()
    _prop_dict: Optional[Dict] = PrivateAttr()
    _g: Optional[Dict] = PrivateAttr()
    _entities: Optional[List] = PrivateAttr()
    _entities_json: Optional[List] = PrivateAttr()

    _imported_ontologies: Optional[List[str]] = PrivateAttr()

    @staticmethod
    def _recursive_ontology_import(
        file: str,
        format_: str = "turtle",
        imported_ontologies: list = None,
        import_mapping=None,
    ):
        """Recursively import ontologies from owl:imports statements in the given
        ontology file.

        Parameters:
        -----------
        file:
            path or url to the ontology file
        format_:
            the serialization format (see rdflib docs)
        imported_ontologies:
            list of already imported ontologies

        Returns:
        --------
            graph (rdflib.Graph): graph of the imported ontologies"""
        if imported_ontologies is None:
            imported_ontologies = []
        if import_mapping is None:
            import_mapping = {}
        rdf_graph = Graph()
        rdf_graph.parse(file, format=format_)

        # load the graph into a dict
        _g = json.loads(rdf_graph.serialize(format="json-ld", auto_compact=True))

        ontologies_to_import = []

        def handle_node(node):
            if "owl:imports" in node:
                if not isinstance(node["owl:imports"], list):
                    node["owl:imports"] = [node["owl:imports"]]
                for import_ in node["owl:imports"]:
                    onto = import_["@id"]
                    if onto not in imported_ontologies:
                        if onto in import_mapping:
                            ontologies_to_import.append(import_mapping[onto])
                        else:
                            ontologies_to_import.append(onto)
                        imported_ontologies.append(onto)
                        # print(node["owl:imports"]["@id"])

        if "@graph" in _g:
            for node in _g["@graph"]:
                handle_node(node)
        else:
            handle_node(_g)

        # load the imported ontologies and merge them into the main graph
        for onto in ontologies_to_import:
            print(f"Importing {onto}")
            _rdf_graph = OntologyImporter._recursive_ontology_import(
                onto,
                format_=format_,
                imported_ontologies=imported_ontologies,
                import_mapping=import_mapping,
            )
            rdf_graph += _rdf_graph

        return rdf_graph

    def import_ontology(self, config: ImportConfig):
        """Imports an ontology into OSW

        Parameters
        ----------
        config
            see ImportConfig

        Returns
        -------
            None
        """

        # overwrite the default document loader to load relative context from the wiki
        jsonld.set_document_loader(self.osw.site.get_jsonld_context_loader())

        self.import_config = config

        # load the ontology file
        self._imported_ontologies = []
        rdf_graph = self._recursive_ontology_import(
            self.import_config.file,
            self.import_config.file_format,
            self._imported_ontologies,
            self.import_config.import_mapping,
        )

        if self.import_config.dump_files:
            rdf_graph.serialize(
                destination=os.path.join(
                    self.import_config.dump_path,
                    f"{self.import_config.ontology_name}.jsonld",
                ),
                format="json-ld",
                auto_compact=True,
            )

        # load the graph into a dict
        self._g = json.loads(rdf_graph.serialize(format="json-ld", auto_compact=True))

        def node_sort(n):
            # sort nodes by type and id
            rank = 10
            if "@type" in n:
                if "Property" in n["@type"]:
                    rank = 0
                if "Class" in n["@type"]:
                    rank = 1
            return (rank, n["@id"])

        self._g["@graph"] = sorted(self._g["@graph"], key=lambda n: node_sort(n))

        # create an id-index of all nodes
        self._id_dict = {}
        for node in self._g["@graph"]:
            if "@id" in node:
                self._id_dict[node["@id"]] = node

        # add a helper context
        self._g["@context"]["_temp_"] = "http://temp.local/"

        # normalize graph
        self._sanitize_graph()

        if self.import_config.dump_files:
            with open(
                os.path.join(
                    self.import_config.dump_path,
                    f"{self.import_config.ontology_name}.mod.jsonld",
                ),
                "w",
                encoding="utf-8",
            ) as f:
                json.dump(self._g, f, indent=4, ensure_ascii=False)

        self._construct_context()
        self._g = jsonld.compact(self._g, self._context)

        # build an iri-index of all nodes
        # ToDo: can we reuse the id-index?
        self._iri_dict = {}
        for node in self._g["@graph"]:
            self._iri_dict[node["iri"]] = node

        self._apply_osw_structure()

        if self.import_config.dump_files:
            with open(
                os.path.join(
                    self.import_config.dump_path,
                    f"{self.import_config.ontology_name}.compacted.jsonld",
                ),
                "w",
                encoding="utf-8",
            ) as f:
                json.dump(self._g, f, indent=4, ensure_ascii=False)

            # currently does not work with relative remote context ("/wiki/..."")
            # rdf_graph.parse(
            #     os.path.join(
            #         self.import_config.dump_path,
            #         f"{self.import_config.ontology_name}.compacted.jsonld"
            #     )
            # )
            # rdf_graph.serialize(
            #     destination=os.path.join(
            #         self.import_config.dump_path,
            #         f"{self.import_config.ontology_name}.jsonld.ttl"
            #     ),
            #     format="ttl"
            # )

        self._create_entities()
        # sort list by type and iri to persist the order across runs
        # self._entities = sorted(self._entities, key=lambda e: (e.type[0], e.iri))

        if self.import_config.dump_files:
            with open(
                os.path.join(
                    self.import_config.dump_path,
                    f"{self.import_config.ontology_name}.entities.jsonld",
                ),
                "w",
                encoding="utf-8",
            ) as f:
                json.dump(self._entities_json, f, indent=4, ensure_ascii=False)

        # Import ontologies
        # self._import_ontology(
        #     OSW.ImportOntologyParam(
        #         ontologies=self.import_config.ontologies,
        #         entities=self._entities,
        #         dryrun=False
        #     )
        # )

        if not self.import_config.dry_run:
            self._store_ontologies(
                OntologyImporter.StoreOntologiesParam(
                    ontologies=self.import_config.ontologies,
                    entities=self._entities,
                )
            )

    def _sanitize_graph(self) -> None:
        ps = self.parser_settings
        # sanitize types of type object, see: https://github.com/RDFLib/rdflib/issues/1950
        for node in self._g["@graph"]:
            key = "@type"
            if key in node:
                # normalize to list
                if not isinstance(node[key], list):
                    node[key] = [node[key]]
                types = []
                for _i, val in enumerate(node[key]):
                    if isinstance(val, dict):
                        if "@id" in val:
                            types.append(val["@id"])
                    else:
                        types.append(val)
                node[key] = types
            # remove 'www.' from @id
            node["@id"] = node["@id"].replace("www.", "")

            # handle subClassOf and subPropertyOf
            if "rdfs:subClassOf" in node:
                if not isinstance(node["rdfs:subClassOf"], list):
                    node["rdfs:subClassOf"] = [node["rdfs:subClassOf"]]
                for i, val in enumerate(node["rdfs:subClassOf"]):
                    id = val["@id"]
                    if id in self._id_dict:
                        superclass = self._id_dict[id]
                        if (
                            "@type" in superclass
                            and superclass["@type"] == "owl:Restriction"
                            and "owl:onProperty" in superclass
                            and "owl:someValuesFrom" in superclass
                        ):
                            if ps.flatten_subclassof_restrictions:
                                if superclass["owl:onProperty"]["@id"] not in node:
                                    node[superclass["owl:onProperty"]["@id"]] = []
                                node[superclass["owl:onProperty"]["@id"]].append(
                                    superclass["owl:someValuesFrom"]["@id"]
                                )
                            else:
                                if (
                                    "_temp_:" + ps.subclassof_restrictions_keyword
                                    not in node
                                ):
                                    node[
                                        "_temp_:" + ps.subclassof_restrictions_keyword
                                    ] = []
                                node[
                                    "_temp_:" + ps.subclassof_restrictions_keyword
                                ].append(superclass)
                            node["rdfs:subClassOf"].pop(i)
                            if id in self._g["@graph"]:
                                self._delete_node(self._g["@graph"], id)
                        # else:
                        #    print(
                        #    "Warning: rdfs:subClassOf is not a Restriction: ",
                        #    superclass
                        #    )

    def _delete_node(self, id: str, key: str = "@id") -> bool:
        """Delete a node from a graph by id

        Parameters
        ----------
        graph
            the graph (list of nodes) to delete from
        id
            the id of the node to delete
        key
            the id key, by default '@id'

        Returns
        -------
            True if the node was deleted, False otherwise
        """
        graph = self._g["@graph"]
        for i, node in enumerate(graph):
            if node[key] == id:
                del graph[i]
                return True
        return False

    def _construct_context(self):
        self._context = {
            "_temp_": "http://temp.local/",
            "owl": "http://www.w3.org/2002/07/owl#",
            "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "xsd": "http://www.w3.org/2001/XMLSchema#",
            "skos": "http://www.w3.org/2004/02/skos/core#",
            "dc": "http://purl.org/dc/terms/",
            # "emmo": "http://emmo.info/emmo#", #keep values with full iri
            "iri": {"@id": "@id"},
            "rdf_type": {"@id": "@type"},
            "label": {"@id": "rdfs:label"},
            # "label": {"@id": "skos:prefLabel"},
            "altLabel": {"@id": "skos:altLabel"},
            "text": {"@id": "@value"},
            "lang": {"@id": "@language"},
            # "subClassOf": {"@id": "rdfs:subClassOf", "@type": "@id"},
            "subclass_of": {"@id": "rdfs:subClassOf", "@type": "@id"},
            "source": "dc:source",
            "disjointUnionOf": "owl:disjointUnionOf",
            "disjointWith": "owl:disjointWith",
            "equivalentClass": "owl:equivalentClass",
            "unionOf": {"@id": "owl:unionOf", "@container": "@list", "@type": "@id"},
            "comment": "rdfs:comment",
            "isDefinedBy": "rdfs:isDefinedBy",
            "seeAlso": "rdfs:seeAlso",
            # shorten properties
            "description": "http://example.com/R7k3ssL7gUxsfWuVsXWDXYF",
            # "ObjectPropertyA": "http://example.com/REh2qNSARmKpPuwrJmr5Pu",
            "Item": "https://wiki-dev.open-semantic-lab.org/wiki",
            # "Item:OSW4db96fb7e9e0466d942bf6f17bfdc145":
            #   "emmo:EMMO_4db96fb7_e9e0_466d_942b_f6f17bfdc145",
            # "uuid": "wiki:HasUuid",
            # "name": "wiki:HasName",
            # "restrictions": {"@id": "rdfs:subClassOf", "@type": "@id"},
            "restrictions": {"@id": "_temp_:restrictions", "@type": "@id"},
            "on_property": {"@id": "owl:onProperty", "@type": "@id"},
            "some_values_from": {"@id": "owl:someValuesFrom", "@type": "@id"},
            "domain": {"@id": "rdfs:domain", "@type": "@id"},
            "range": {"@id": "rdfs:range", "@type": "@id"},
            "subproperty_of": {"@id": "rdfs:subPropertyOf", "@type": "@id"},
            # "example": "http://example.com/",
        }

        # base_class_page = self.osw.site.get_page(WtSite.GetPageParam(
        #    titles=[self.import_config.base_class_title]
        # )).pages[0]
        # schema = base_class_page.get_slot_content("jsonschema")
        # if (isinstance(schema, str)):
        #    schema = json.loads(schema)
        # self._context = schema["@context"]

        self._context = [
            f"/wiki/{self.import_config.base_class_title}?action=raw&slot=jsonschema&format=json",
            {
                "_temp_": "http://temp.local/",
                "restrictions": {"@id": "_temp_:restrictions", "@type": "@id"},
            },
        ]

    def _get_uuid_from_iri(self, iri):
        node_name = iri.split("/")[-1]
        if "#" in node_name:
            node_name = node_name.split("#")[-1]
        try:
            # try to get last 32 hex digits and convert it to uuid
            # e. g. http://emmo.info/emmo/EMMO_4db96fb7_e9e0_466d_942b_f6f17bfdc145
            node_uuid = uuid.UUID(re.sub(r"[^A-Fa-f0-9]", "", node_name)[-32:])
        except Exception:
            node_uuid = uuid.uuid5(uuid.NAMESPACE_URL, iri)
        return node_uuid

    def _get_page_name(self, node):
        ns = ""
        title = ""

        node_uuid = self._get_uuid_from_iri(node["iri"])

        if "rdf_type" in node:
            if "owl:Class" in node["rdf_type"]:
                ns = "Category"
                title = OSW.get_osw_id(node_uuid)
            elif (
                "owl:ObjectProperty" in node["rdf_type"]
                or "owl:DatatypeProperty" in node["rdf_type"]
                or "owl:AnnotationProperty" in node["rdf_type"]
            ):
                ns = "Property"
                if self.import_config.property_naming_policy == "label":
                    title = node["name"]
                elif self.import_config.property_naming_policy == "prefixed_label":
                    prefix = None
                    for onto in self.import_config.ontologies:
                        if (
                            onto.iri in node["iri"]
                            or onto.prefix_name
                            + self.import_config.property_naming_prefix_delimiter
                            in node["iri"]
                        ):
                            prefix = onto.prefix_name
                            break

                    if prefix:
                        name = None
                        if "name" in node:
                            name = node["name"]
                        elif ":" in node["iri"]:
                            name = node["iri"].split(":")[-1]
                        if name:
                            title = (
                                f"{prefix}"
                                f"{self.import_config.property_naming_prefix_delimiter}"
                                f"{name}"
                            )
                        else:
                            raise ValueError(
                                f"Could not find name for property {node['iri']}"
                            )
                    else:
                        raise ValueError(
                            f"Could not find prefix for property {node['iri']}"
                        )
                if self.import_config.property_naming_policy == "uuid":
                    title = OSW.get_osw_id(node_uuid)
            elif "owl:NamedIndividual" in node["rdf_type"]:
                ns = "Item"
                title = OSW.get_osw_id(node_uuid)
        else:  # e. g. owl:Axiom
            return None

        return f"{ns}:{title}"

    def _map_iri_to_osw(self, iri):
        result = None
        if iri in self._iri_dict:
            node = self._iri_dict[iri]
            result = self._get_page_name(node)
        if not result:
            result = iri
        return result

    def _apply_osw_structure(self):
        ps = self.parser_settings
        for node in self._g["@graph"]:
            for key in ps.ensure_multilang:
                if key in node:
                    if isinstance(node[key], str):
                        node[key] = {"text": node[key], "lang": "en"}
                    elif "text" in node[key] and "lang" not in node[key]:
                        node[key]["lang"] = "en"
                    elif isinstance(node[key], list):
                        for i, _val in enumerate(node[key]):
                            if isinstance(node[key][i], str):
                                node[key][i] = {"text": node[key][i], "lang": "en"}
                            elif "text" in node[key][i] and "lang" not in node[key][i]:
                                node[key][i]["lang"] = "en"
            for key in ps.ensure_array:
                if key in node and not isinstance(node[key], list):
                    node[key] = [node[key]]
            for key in ps.remove_unnamed:
                if key in node:
                    if isinstance(node[key], list):
                        node[key] = [
                            value for value in node[key] if not value.startswith("_:")
                        ]
                    elif isinstance(node[key], str) and node[key].startswith("_:"):
                        del node[key]

            if "rdf_type" in node and any(
                x in node["rdf_type"]
                for x in [
                    "owl:Class",
                    "owl:ObjectProperty",
                    "owl:DatatypeProperty",
                    "owl:AnnotationProperty",
                    "owl:NamedIndividual",
                ]
            ):
                node["uuid"] = str(self._get_uuid_from_iri(node["iri"]))

                # name = ""
                if "prefLabel" in node:
                    # if isinstance(node["prefLabel"], str):
                    #   node["prefLabel"] = {"text": node["prefLabel"], "lang": "en"}
                    # if isinstance(node["prefLabel"], list):
                    #   node["name"] = node["prefLabel"][0]["text"]
                    # else: node["name"] = node["prefLabel"]["text"]
                    node["name"] = node["prefLabel"][0]["text"]
                elif "label" in node:
                    node["name"] = node["label"][0]["text"]
                else:
                    print("No label: ", node["iri"])
                if "name" in node:
                    node["name"] = pascal_case(node["name"])

                # update iri dict
                self._iri_dict[node["iri"]] = node

                namespace = self._get_page_name(node).split(":")[0]
                title = self._get_page_name(node).replace(namespace + ":", "")
                node["meta"] = {
                    "wiki_page": {
                        "title": title,
                        "namespace": namespace,
                    }
                }

                for onto in self.import_config.ontologies:
                    if (
                        onto.prefix in node["iri"]
                        or onto.prefix_name + ":" in node["iri"]
                    ):
                        node["imported_from"] = (
                            onto.prefix_name
                            + ":"
                            + node["iri"]
                            .replace(onto.prefix, "")
                            .replace(onto.prefix_name + ":", "")
                        )
                        break

        self._map_iris(self._g["@graph"])

        # build an inverted property_dict
        self._prop_dict = {}
        for node in self._g["@graph"]:
            self._iri_dict[node["iri"]] = node
        for node in self._g["@graph"]:
            for key in node:
                if ":" in key and key not in self._prop_dict:
                    label = ""
                    if key in self._iri_dict and "label" in self._iri_dict[key]:
                        label = self._iri_dict[key]["label"][0]["text"]
                    self._prop_dict[key] = label
        # print the inverted property_dict
        for key, value in self._prop_dict.items():
            print(f'"{value}": "{key}"')

    def _map_iris(self, node_array):
        if not isinstance(node_array, list):
            node_array = [node_array]
        for node in node_array:
            if isinstance(node, dict):
                for key in node:
                    if key in self.parser_settings.map_uuid_iri:
                        if isinstance(node[key], list):
                            for i, _val in enumerate(node[key]):
                                node[key][i] = self._map_iri_to_osw(node[key][i])
                        if isinstance(node[key], str):
                            node[key] = self._map_iri_to_osw(node[key])
                    if key == "restrictions":
                        print(key)
                    if isinstance(node[key], dict) or isinstance(node[key], list):
                        self._map_iris(node[key])

    def _create_entities(self):
        # create OSW entities
        limit = 3000  # choose a smaller number for tests
        counter = 0
        self._entities = []
        self._entities_json = []
        for _index, node in enumerate(self._g["@graph"]):
            # restore class restrictions
            # if "_restrictions" in node:
            #    node["restrictions"] == node["_restrictions"]

            if counter < limit:
                if "rdf_type" in node and "label" in node:
                    e = None
                    if "owl:Class" in node["rdf_type"]:
                        e = self.import_config.base_class(**node)

                    if "owl:ObjectProperty" in node["rdf_type"]:
                        e = model.ObjectProperty(**node)

                    if "owl:DatatypeProperty" in node["rdf_type"]:
                        e = model.DataProperty(**node)

                    if "owl:AnnotationProperty" in node["rdf_type"]:
                        e = model.AnnotationProperty(**node)

                    if "owl:NamedIndividual" in node["rdf_type"]:
                        e = model.OwlIndividual(**node)

                    if e:
                        self._entities.append(e)
                        self._entities_json.append(
                            json.loads(e.json(exclude_none=True))
                        )
                        counter += 1

    class StoreOntologyParam(model.OswBaseModel):
        ontology: model.OwlOntology
        entities: List[model.OswBaseModel]

    def _store_ontology(self, param: StoreOntologyParam):
        import_page = self.osw.site.get_page(
            WtSite.GetPageParam(
                titles=["MediaWiki:Smw_import_" + param.ontology.prefix_name]
            )
        ).pages[0]
        text = f"{param.ontology.prefix} | [{param.ontology.see_also[0]} {param.ontology.name}]"
        for e in param.entities:
            namespace = get_namespace(e)

            # see https://www.semantic-mediawiki.org/wiki/Help:Import_vocabulary
            if namespace == "Category":
                smw_import_type = "Category"
                if not hasattr(e, "subclass_of"):
                    e.subclass_of = []
                if len(e.subclass_of) == 0:
                    e.subclass_of.append(self.import_config.meta_class_title)
            elif namespace == "Property":
                smw_import_type = "Type:" + e.cast(model.Property).property_type
            else:
                smw_import_type = "Item"
            iri = None
            if hasattr(e, "iri"):
                iri = e.iri
            if hasattr(e, "uri"):
                iri = e.uri
            if iri is not None:
                new_iri = iri.replace(param.ontology.prefix, "").replace(
                    param.ontology.prefix_name + ":", ""
                )
                text += f"\n {new_iri}|{smw_import_type}"
            else:
                print("Error: Entity has not iri/uri property")

        import_page.set_slot_content("main", text)
        import_page.edit("import ontology")

        self.osw.store_entity(OSW.StoreEntityParam(entities=param.entities))

    @deprecated("use ontology.OntologyImporter.StoreOntologiesParam instead")
    class StoreOntologiesParam(model.OswBaseModel):
        entities: Optional[List[model.OswBaseModel]]
        """If we use model.Entity here, all instances are casted to model.Entity
        See: https://stackoverflow.com/questions/67366187/
        how-to-make-a-pydantic-field-accept-subclasses-using-type
        """
        # classes: Optional[List[model.Entity]]  # goes to NS Category
        # properties: Optional[List[model.Entity]]  # goes to NS Property
        # individuals: Optional[List[model.Entity]]  # goes to NS Item
        ontologies: Optional[List[model.OwlOntology]]
        dryrun: Optional[bool] = False

    def _store_ontologies(self, param: StoreOntologiesParam = None):
        entities = self._entities
        dry_run = self.import_config.dry_run
        ontologies = self.import_config.ontologies
        if param:
            entities = param.entities
            dry_run = param.dryrun
            ontologies = param.ontologies

        print(f"{len(entities)} classes to import")
        prefix_dict = {}
        for e in entities:
            if "//" in e.iri:
                if "http://purl.obolibrary.org/obo/" in e.iri:
                    # special handling for OBO ontologies
                    # ontologies are distinguished by the first part of the postfix
                    # e. g. http://purl.obolibrary.org/obo/BFO_0000050
                    postfix = e.iri.replace("http://purl.obolibrary.org/obo/", "")
                    key = (
                        "http://purl.obolibrary.org/obo/" + postfix.split("_")[0] + "_"
                    )
                elif "#" in e.iri:
                    key = e.iri.split("#")[0] + "#"
                else:
                    key = e.iri.replace(e.iri.split("/")[-1], "")
            else:
                key = e.iri.split(":")[0]  # e. g. "obo:OBO_0000050"
            if key not in prefix_dict:
                prefix_dict[key] = []
            prefix_dict[key].append(e)

        for prefix in prefix_dict.keys():
            onto = None
            for o in ontologies:
                if o.prefix == prefix or o.prefix_name == prefix:
                    onto = o
            if onto is None:
                print(f"Error: No ontology defined for prefix {prefix}")
            else:
                if not dry_run:
                    self._store_ontology(
                        OntologyImporter.StoreOntologyParam(
                            ontology=onto, entities=prefix_dict[prefix]
                        )
                    )



"""This script provides a GUI for downloading an OSW instances page slots, editing
them locally and uploading them again."""

import json
import os

# import yaml
from pathlib import Path
from typing import Union

import PySimpleGUI as psg
from numpy import array as np_array

import osw.model.page_package as package
from osw.auth import CredentialManager
from osw.express import OswExpress
from osw.wiki_tools import read_domains_from_credentials_file
from osw.wtsite import SLOTS, WtPage, WtSite

# Definition of constants
GUI_THEME = "reddit"
DEBUG = True  # Set to True to get an output element in the GUI
DUMP_EMPTY_SLOTS_DEFAULT = True
SEL_INDICES_DEFAULT = [
    0,  # "main",
    3,  # "jsondata",
    4,  # "jsonschema",
    6,  # "header_template",
]
SLOTS_TO_UPLOAD_DEFAULT = np_array(list(SLOTS.keys()))[SEL_INDICES_DEFAULT].tolist()
TARGET_PAGE_DEFAULT = "https://wiki-dev.open-semantic-lab.org/wiki/Main_Page"
LWD_DEFAULT = Path(os.getcwd()).parent / "data"
SETTINGS_FILE_PATH_DEFAULT = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "settings.json"
)
CRED_FILEPATH_DEFAULT = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
PACKAGE_INFO_FILE_NAME = "packages.json"
# Sub folder to store the slot content of the page in. If None, the content is stored
#  in the root folder alongside the package info file.
SUB_LEVEL = None  # "content"
# Whether to store the page content in a sub folder named after the namespace
NAMESPACE_AS_FOLDER = False


# Definition of functions
def str_or_none(value):
    if str(value) == "" or value is None:
        return str("None")
    else:
        return str(value)


def create_config_from_setting(settings_: dict) -> WtPage.PageDumpConfig:
    config_ = WtPage.PageDumpConfig(
        target_dir=settings_["local_working_directory"],
        namespace_as_folder=settings_["namespace_as_folder"],
        dump_empty_slots=settings_["dump_empty_slots"],
        page_name_as_filename=settings_["page_name_as_filename"],
    )
    return config_


def save_as_page_package(
    full_page_name_str,
    wtsite_inst: WtSite,
    dump_config_inst: WtPage.PageDumpConfig,
    label_str: str = None,
    top_level: str = None,
    sub_level: str = None,  # "content"
    author: Union[str, list] = "Open Semantic World",
):
    if top_level is None:
        top_level = full_page_name_str.split(":")[-1]
    package_repo_org = "OpenSemanticWorld-Packages"
    package_repo = f"{top_level}.{sub_level}"
    package_id = f"{top_level}.{sub_level}"
    package_name = top_level
    package_subdir = sub_level
    package_branch = "deleteme"
    publisher = "Open Semantic World"
    working_dir = dump_config_inst.target_dir
    if dump_config_inst.page_name_as_filename and label_str is not None:
        target_dir = os.path.join(working_dir, label_str)
    else:
        target_dir = os.path.join(working_dir, top_level)
    if isinstance(author, list):
        author_list = author
    elif isinstance(author, str):
        author_list = [author]
    else:
        author_list = [""]
    if package_subdir is None:
        base_url = (
            f"https://raw.githubusercontent.com/{package_repo_org}/"
            f"{package_repo}/{package_branch}/"
        )
        content_path = target_dir
    else:
        base_url = (
            f"https://raw.githubusercontent.com/{package_repo_org}/"
            f"{package_repo}/{package_branch}/{package_subdir}/"
        )
        content_path = os.path.join(target_dir, package_subdir)
    bundle = package.PagePackageBundle(
        publisher=publisher,
        author=author_list,
        language="en",
        publisherURL=f"https://github.com/{package_repo_org}/{package_repo}",
        packages={
            f"{package_name}": package.PagePackage(
                globalID=f"{package_id}",
                label=package_name,
                version="0.0.1",
                description="Created by the GUI for local editing of OSW pages",
                baseURL=base_url,
            )
        },
    )
    config = package.PagePackageConfig(
        name=package_name,
        config_path=os.path.join(target_dir, PACKAGE_INFO_FILE_NAME),
        content_path=content_path,
        bundle=bundle,
        titles=[full_page_name_str],
    )
    wtsite_inst.create_page_package(
        WtSite.CreatePagePackageParam(
            config=config,
            dump_config=dump_config_inst,
            debug=False,
        )
    )
    return {"Page package bundle": bundle, "Page package config": config}


if __name__ == "__main__":
    # Predefining some variables before execution
    settings_file_path = SETTINGS_FILE_PATH_DEFAULT
    if os.path.exists(settings_file_path):
        with open(settings_file_path, "r") as f:
            settings = json.load(f)
        settings_read_from_file = True
    else:
        settings = {
            "cred_filepath": str(CRED_FILEPATH_DEFAULT),
            "local_working_directory": str(LWD_DEFAULT),
            "settings_file_path": str(SETTINGS_FILE_PATH_DEFAULT),
            "target_page": TARGET_PAGE_DEFAULT,
            "namespace_as_folder": NAMESPACE_AS_FOLDER,
            "dump_empty_slots": DUMP_EMPTY_SLOTS_DEFAULT,
            "page_name_as_filename": True,
            "slots_to_upload": SLOTS_TO_UPLOAD_DEFAULT,
            "domain": "",
        }
        settings_read_from_file = False

    domains, accounts = read_domains_from_credentials_file(settings["cred_filepath"])
    if "wiki-dev.open-semantic-lab.org" in domains:
        domain = "wiki-dev.open-semantic-lab.org"
    else:
        domain = domains[0]
    if settings_read_from_file:
        settings["domain"] = domain

    cm = CredentialManager(cred_filepath=settings["cred_filepath"])
    osw_obj = OswExpress(domain=domain, cred_mngr=cm)
    wtsite_obj = osw_obj.site

    full_page_name = settings["target_page"].split("/")[-1].replace("_", " ")
    page = wtsite_obj.get_page(WtSite.GetPageParam(titles=[full_page_name])).pages[0]
    label_set = False
    label = None
    slots_downloaded = False
    dump_config = create_config_from_setting(settings)

    # ----- GUI Definition -----
    # Setting the theme of the GUI
    psg.theme(GUI_THEME)
    # Settings
    settings_layout = [
        [psg.Text("Settings", font=("Helvetica", 20))],
        [
            psg.Column(
                [
                    [
                        psg.Text("Settings file"),
                    ],
                    [
                        psg.Text("Credentials file"),
                    ],
                    [
                        psg.Text("Local working directory"),
                    ],
                ]
            ),
            psg.Column(
                [
                    [
                        psg.In(
                            size=(50, 1),
                            enable_events=True,
                            key="-SETTINGS-",
                            default_text=settings["settings_file_path"],
                        ),
                        psg.FileBrowse(button_text="Browse", key="-BROWSE_SETTINGS-"),
                        psg.Button("Load", key="-LOAD_SETTINGS-"),
                        psg.Button("Save", key="-SAVE_SETTINGS-"),
                    ],
                    [
                        psg.In(
                            size=(50, 1),
                            enable_events=True,
                            key="-CREDENTIALS-",
                            default_text=settings["cred_filepath"],
                        ),
                        psg.FileBrowse(
                            button_text="Browse", key="-BROWSE_CREDENTIALS-"
                        ),
                    ],
                    [
                        psg.In(
                            size=(50, 1),
                            enable_events=True,
                            key="-LWD-",
                            default_text=settings["local_working_directory"],
                        ),
                        psg.FolderBrowse(button_text="Browse", key="-BROWSE_LWD-"),
                    ],
                ]
            ),
        ],
        [psg.HSeparator()],
    ]
    # Actions
    actions_layout = [
        # Actions
        [psg.Text("Actions", font=("Helvetica", 20))],
        # Target OSW instance
        [psg.Text("Target OSW instance", font=("Helvetica", 16))],
        [psg.Text("List of domains is read from accounts.pwd.yaml!")],
        [psg.Combo(domains, default_value=domain, key="-DOMAIN-", enable_events=True)],
        # Target page
        [psg.Text("Target page", font=("Helvetica", 16))],
        [
            psg.Column(
                [
                    [psg.Text("Address, within selected OSW instance")],
                    [psg.Text("First label of the OSW page:")],
                ]
            ),
            psg.Column(
                [
                    [
                        psg.InputText(
                            size=(50, 1),
                            default_text=settings["target_page"],
                            key="-ADDRESS-",
                        ),
                        psg.Button("Load page"),
                    ],
                    [
                        # A display element that will show the label of the OSW page
                        psg.Multiline(size=(50, 1), key="-LABEL-", no_scrollbar=True)
                    ],
                ]
            ),
        ],
        # [
        #     psg.HSeparator()
        # ],
        # Slots to download
        [psg.Text("Slots to download", font=("Helvetica", 16))],
        [
            psg.Column(
                [
                    [
                        psg.Radio(
                            "Include empty slots",
                            group_id="-RADIO1-",
                            default=settings["dump_empty_slots"],
                            key="-INC_EMPTY-",
                            enable_events=True,
                        )
                    ],
                    [
                        psg.Radio(
                            "Exclude empty slots",
                            group_id="-RADIO1-",
                            default=(not settings["dump_empty_slots"]),
                            key="-EXC_EMPTY-",
                            enable_events=True,
                        )
                    ],
                    [psg.Button("Download selected", key="-DL-")],
                    [psg.Multiline(size=(20, 1), key="-DL_RES-", no_scrollbar=True)],
                ]
            ),
        ],
        # [
        #     psg.HSeparator()
        # ],
        # Slots to upload
        [psg.Text("Slots to upload", font=("Helvetica", 16))],
        [
            psg.Column(
                [
                    [psg.Button("(De)Select all", key="-UL_SEL-")],
                    [
                        psg.Listbox(
                            values=SLOTS.keys(),
                            default_values=settings["slots_to_upload"],
                            size=(20, 10),
                            key="-UL_LIST-",
                            select_mode=psg.LISTBOX_SELECT_MODE_MULTIPLE,
                            enable_events=True,
                            no_scrollbar=True,
                        )
                    ],
                    [psg.Button("Upload selected", key="-UL-")],
                    [psg.Multiline(size=(20, 1), key="-UL_RES-", no_scrollbar=True)],
                ]
            )
        ],
    ]
    output_layout = [
        [psg.HSeparator()],
        # Output
        [psg.Text("Debug output", font=("Helvetica", 20))],
        [psg.Multiline(key="-OUTPUT-", size=(100, 10))],
    ]
    # The full layout
    layout = settings_layout + actions_layout
    if DEBUG:
        layout += output_layout
    # Create the window
    window = psg.Window("OSW local editing GUI", layout)
    # Create an event loop
    while True:
        event, values = window.read()
        # End program if user closes window or
        # presses the OK button
        if event == "OK" or event == psg.WIN_CLOSED:
            break
        elif event == "-SETTINGS-":
            settings["settings_file_path"] = values["-SETTINGS-"]
        elif event == "-LOAD_SETTINGS-":
            # update settings
            with open(settings["settings_file_path"], "r") as f:
                settings = json.load(f)
            # update GUI
            window["-CREDENTIALS-"].update(settings["cred_filepath"])
            window["-LWD-"].update(settings["local_working_directory"])
            window["-DOMAIN-"].update(settings["domain"])
            window["-ADDRESS-"].update(settings["target_page"])
            window["-INC_EMPTY-"].update(settings["dump_empty_slots"])
            window["-EXC_EMPTY-"].update(not settings["dump_empty_slots"])
            indices = [
                i
                for i, x in enumerate(SLOTS.keys())
                if x in settings["slots_to_upload"]
            ]
            window["-UL_LIST-"].update(set_to_index=indices)
            settings_read_from_file = True
        elif event == "-SAVE_SETTINGS-":
            with open(settings["settings_file_path"], "w") as f:
                json.dump(settings, f, indent=4)
        elif event == "-CREDENTIALS-":
            settings["cred_filepath"] = values["-CREDENTIALS-"]
            domains, accounts = read_domains_from_credentials_file(
                settings["cred_filepath"]
            )
            window["-DOMAIN-"].update(values=domains)
        elif event == "-LWD-":
            settings["local_working_directory"] = values["-LWD-"]
        elif event == "-DOMAIN-":
            settings["domain"] = values["-DOMAIN-"]
            domain = settings["domain"].split("//")[-1]
            osw_obj = OswExpress(domain=settings["domain"], cred_mngr=cm)
            wtsite_obj = osw_obj.site
        elif event == "Load page":
            window["-LABEL-"].update("Loading page...")
            window["-DL_RES-"].update("")
            full_page_name = values["-ADDRESS-"].split("/")[-1].replace("_", " ")
            if (values["-ADDRESS-"].find("/wiki/") != -1) or (
                values["-ADDRESS-"].find("/w/") != -1
            ):
                settings["target_page"] = values["-ADDRESS-"]
            else:
                settings["target_page"] = (
                    "https://" + domain + "/wiki/" + full_page_name
                )
            if values["-ADDRESS-"].find(settings["domain"]) == -1:
                window["-LABEL-"].update("Page not on selected domain!")
                label_set = False
            else:
                # use connection
                page = wtsite_obj.get_page(
                    WtSite.GetPageParam(titles=[full_page_name])
                ).pages[0]
                if page.exists:
                    jsondata = page.get_slot_content("jsondata")
                    if jsondata is None:
                        window["-LABEL-"].update("Slot 'jsondata' is empty!")
                        label_set = False
                    else:
                        label = jsondata["label"][0]["text"]
                        window["-LABEL-"].update(label)
                        label_set = True
                else:
                    window["-LABEL-"].update("Page does not exist!")
                    label_set = False
        elif event == "-EXC_EMPTY-" or event == "-INC_EMPTY-":
            settings["dump_empty_slots"] = values["-INC_EMPTY-"]
        elif event == "-DL-":
            window["-DL_RES-"].update("Downloading slots...")
            if label_set:
                dump_config = create_config_from_setting(settings)
                _ = save_as_page_package(
                    full_page_name_str=full_page_name,
                    wtsite_inst=wtsite_obj,
                    dump_config_inst=dump_config,
                    label_str=label,
                    sub_level=SUB_LEVEL,
                    author=accounts[domains[0]]["username"],
                )
                slots_downloaded = True
                window["-DL_RES-"].update("Slots downloaded!")
            else:
                window["-DL_RES-"].update("No page loaded!")
        elif event == "-UL_SEL-":
            indices = list(window["-UL_LIST-"].get_indexes())
            # All selected: deselect all
            if indices == list(range(len(SLOTS.keys()))):
                window["-UL_LIST-"].update(set_to_index=[])
                settings["slots_to_upload"] = []
            else:
                window["-UL_LIST-"].update(set_to_index=list(range(len(SLOTS.keys()))))
                settings["slots_to_upload"] = list(SLOTS.keys())
        elif event == "-UL_LIST-":
            indices = list(window["-UL_LIST-"].get_indexes())
            settings["slots_to_upload"] = np_array(list(SLOTS.keys()))[indices].tolist()
        elif event == "-UL-":
            slots_to_upload = settings["slots_to_upload"]
            if not label_set:
                window["-UL_RES-"].update("No page loaded!")
            elif len(slots_to_upload) == 0:
                window["-UL_RES-"].update("No slots selected!")
            else:
                if not slots_downloaded:
                    window["-UL_RES-"].update("No slots downloaded!")
                    dump_config = create_config_from_setting(settings)
                window["-UL_RES-"].update("Uploading slots...")
                if SUB_LEVEL is None:
                    storage_path = Path(dump_config.target_dir)
                else:
                    storage_path = Path(dump_config.target_dir).parent
                pages = wtsite_obj.read_page_package(
                    WtSite.ReadPagePackageParam(
                        storage_path=storage_path,
                        packages_info_file_name=PACKAGE_INFO_FILE_NAME,
                        selected_slots=slots_to_upload,
                        debug=False,
                    )
                ).pages
                param = wtsite_obj.UploadPageParam(pages=pages, parallel=False)
                wtsite_obj.upload_page(param)
                # Success:
                window["-UL_RES-"].update("Slots uploaded!")
                # Report in the download section that the slots have been uploaded to
                #  remind the user that he eventually has to re-download the slots
                window["-DL_RES-"].update("Slots uploaded!")

        # Some debugging output functionality
        if DEBUG:
            values_str = "\n".join(
                [
                    f"{key}: {str_or_none(value)}"
                    for key, value in values.items()
                    if key != "-OUTPUT-"
                ]
            )
            listbox_selected_items_str = "\n".join(settings["slots_to_upload"])

            window["-OUTPUT-"].update(
                f"Last event: {event}\n"
                f"-----------\n"
                f"Values:\n"
                f"-------\n"
                f"{values_str}"
                f"\nListbox values:\n"
                f"{listbox_selected_items_str}"
            )

    window.close()



import asyncio
import uuid
from os import environ
from typing import Optional
from uuid import UUID, uuid4

from prefect import flow, get_client, task
from prefect.blocks.system import Secret
from pydantic.v1 import Field

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.core import OSW
from osw.utils.wiki import get_full_title
from osw.wtsite import WtSite


class ConnectionSettings(model.OswBaseModel):
    """Connection data for OSW"""

    osw_user_name: Optional[str]
    """The login username.
    Note: value of envar OSW_USER used of not given
    Note: value of envar OSW_PASSWORD used for login"""
    osw_domain: Optional[str]
    """The domain of the instance
    Note: value of envar OSW_SERVER used of not given"""


@task
def connect(settings: Optional[ConnectionSettings] = None):
    """Initiates the connection to the OSW instance

    Parameters
    ----------
    settings
        see ConnectionSetttings dataclass
    """
    if settings is None:
        settings = ConnectionSettings()
    global wtsite
    # define username
    if environ.get("OSW_USER") is not None and environ.get("OSW_USER") != "":
        settings.osw_user_name = environ.get("OSW_USER")
    if environ.get("OSW_SERVER") is not None and environ.get("OSW_SERVER") != "":
        settings.osw_domain = environ.get("OSW_SERVER")
    password = ""
    if environ.get("OSW_PASSWORD") is not None and environ.get("OSW_PASSWORD") != "":
        password = environ.get("OSW_PASSWORD")
    else:
        # fetch secret stored in prefect server from calculated name
        password = Secret.load(
            settings.osw_user_name.lower() + "-" + settings.osw_domain.replace(".", "-")
        ).get()  # e. g. mybot-wiki-dev-open-semantic-lab-org
    cm = CredentialManager()
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=settings.osw_domain, username=settings.osw_user_name, password=password
        )
    )
    wtsite = WtSite(WtSite.WtSiteConfig(iri=settings.osw_domain, cred_mngr=cm))
    global osw
    osw = OSW(site=wtsite)


@task
def fetch_schema():
    """this will load the current entity schema from the OSW instance."""
    # Load Article Schema on demand
    if not hasattr(model, "Article"):
        osw.fetch_schema(
            OSW.FetchSchemaParam(
                schema_title=[
                    "Category:OSW77e749fc598341ac8b6d2fff21574058",  # Software
                    "Category:OSW72eae3c8f41f4a22a94dbc01974ed404",  # PrefectFlow
                    "Category:OSW92cc6b1a2e6b4bb7bad470dfdcfdaf26",  # Article
                ],
                mode="replace",
            )
        )


class Result(model.OswBaseModel):
    """The result dataclass"""

    uuid: Optional[UUID] = Field(default_factory=uuid4, title="UUID")
    """UUIDv4 of the result (autogenerated)"""
    target_title: Optional[str] = None
    """the target title to store the result.
    Autogenerated from uuid if not given
    """
    msg: str
    """The message you want to leave on the target page"""


@task
def store_and_document_result(result: Result):
    """Store workflow results on an OSW entry

    Parameters
    ----------
    result
        see Result dataclass
    """
    if result.target_title:
        title = result.target_title
    else:
        title = "Item:" + osw.get_osw_id(result.uuid)
    entity = osw.load_entity(title)
    if entity is None:
        # does not exist yet - create a new one
        entity = model.Article(
            uuid=result.uuid, label=[model.Label(text="Article for dummy workflow")]
        )

    # edit structured data
    entity = entity.cast(model.Article)
    entity.description = [model.Description(text="some descriptive text")]
    osw.store_entity(entity)

    # edit unstructured data (ToDo: access page directly from entity)
    page = osw.site.get_page(WtSite.GetPageParam(titles=[title])).pages[0]
    page.set_slot_content(
        slot_key="main", content=page.get_slot_content("main") + "<br>\n" + result.msg
    )
    page.edit()
    print("FINISHED")


class Request(model.OswBaseModel):
    uuid: UUID = Field(default_factory=uuid4, title="UUID")
    """UUIDv4 of the request."""
    osw_domain: Optional[str] = "wiki-dev.open-semantic-lab.org"
    """To domain of the OSW instance"""
    subject: Optional[str] = "Item:OSW56f9439d43244fe7a83163bab9414ee1"
    """Where to store the results. For testing, we use a static default value"""
    msg: Optional[str] = "test message"
    """The message you want to leave on the target page"""


@flow(
    name="Dummy Workflow",
    description="Dummy workflow that prints a message on a target page",
)
def dummy_workflow(request: Request):
    """Dummy workflow that prints a message on a target page

    Parameters
    ----------
    request
        see Request dataclass
    """
    connect(ConnectionSettings(osw_domain=request.osw_domain))
    fetch_schema()
    store_and_document_result(Result(msg=request.msg, target_title=request.subject))


async def deploy():
    """programmatic deployment supported in newer prefect versions"""
    flow = dummy_workflow
    # flow_name = flow.name
    deployment_name = flow.name + " Deployment"

    # create a deployment and apply it
    config = await flow.to_deployment(name=deployment_name)
    await config.apply()  # returns the deployment_uuid

    # fetch flow uuid
    async with get_client() as client:
        response = await client.read_flow_by_name(flow.name)
        print(response.json())
        flow_uuid = response.id

    await connect()
    await fetch_schema()
    # static UUIDv5 namespace for a stable UUID
    namespace_uuid = uuid.UUID("0dd6c54a-b162-4552-bab9-9942ccaf4f41")

    # self-documentation / registration
    this_tool = model.Software(
        uuid=uuid.uuid5(namespace_uuid, flow.name),
        label=[model.Label(text=flow.name)],
        description=[model.Description(text=flow.description)],
    )

    prefect_domain = environ.get("PREFECT_API_URL").split("//")[-1].split("/")[0]
    this_flow = model.PrefectFlow(
        uuid=flow_uuid,
        label=[model.Label(text=flow.name + " Prefect Flow")],
        description=[model.Description(text=flow.description)],
        flow_id=str(flow_uuid),
        hosted_software=[get_full_title(this_tool)],
        domain=prefect_domain,
    )

    osw.store_entity(osw.StoreEntityParam(entities=[this_tool, this_flow]))

    # start agent to serve deployment
    await dummy_workflow.serve(name=deployment_name)


if __name__ == "__main__":
    # dummy_workflow(Request(msg="Test"))
    with asyncio.Runner() as runner:
        runner.run(deploy())



import re
from typing import Dict, List, Optional, Union

from pydantic.v1 import validator

from osw.model.static import OswBaseModel


# Classes
class RegExPatternExtended(OswBaseModel):
    """A class that extends the re.Pattern class with additional attributes.

    Example
    -------
    Definition
    >>> pattern = RegExPatternExtended(
    >>>     description="File page full page title (new format) to label",
    >>>     pattern=r"File:(OS[WL]{1}[[a-f0-9]{32})"
    >>>     "(\\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
    >>>     group_keys=["Label", "Suffix(es)"],  # ["OSW-ID", "Suffix(es)"]
    >>>     example_str="File:OSW11a53cdfbdc24524bf8ac435cbf65d9d.svg",
    >>>     expected_groups=["OSW11a53cdfbdc24524bf8ac435cbf65d9d", ".svg"],
    >>> )
    Usage
    >>> some_str = pattern.example_str
    >>> match = pattern.match(some_str)
    >>> label = match.groups["label"]
    >>> suffix = match.groups["Suffix(es)"]
    Testing
    >>> pattern.test_pattern()


    Resources
    ---------
    https://regex101.com/
    """

    pattern: Union[str, re.Pattern]
    """The regular expression pattern. Can be a string or a compiled pattern."""
    description: str
    """A (unique!) description of the pattern. This is used as a key in a dictionary
    containing all patterns and acting as a library."""
    group_keys: List[str]
    """A list of keys that can be used to access the match groups in the match
    result. The number of keys must match the number of match groups in the pattern.
    So the first key (list index 0) will be used to access the first match group
    (index 1!)."""
    example_str: Optional[str]
    """An example string that can be used to test the pattern."""
    expected_groups: Optional[List[str]]
    """A list strings that are expected in the example_match.groups.values(). This is
    used to test the pattern by asserting list(self.example_match.groups.values()) ==
    self.expected_groups"""

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)
        if isinstance(self.pattern, str):
            self.pattern = re.compile(self.pattern)

    def match(self, string: str) -> "MatchResult":
        """Return the match result of the given string and the pattern."""
        return MatchResult(match=re.match(self.pattern, string), pattern=self)

    def search(self, string: str) -> "SearchResult":
        """Return the search result of the given string and the pattern."""
        return SearchResult(match=re.search(self.pattern, string), pattern=self)

    def finditer(self, string: str) -> List["MatchResult"]:
        """Return a list of all match results of the given string and the pattern."""
        return [
            MatchResult(match=match, pattern=self)
            for match in re.finditer(self.pattern, string)
        ]

    def findall(self, string: str) -> Dict[str, List[str]]:
        """Return a list of all match groups of the given string and the pattern."""
        result = {}
        for grp_key in self.group_keys:
            result[grp_key] = [mr.groups[grp_key] for mr in self.finditer(string)]
        return result

    def findall_by_group_key(self, string: str, group_key: str) -> List[str]:
        """Return a list of all match groups of the given string and the pattern by
        group key."""
        if group_key not in self.group_keys:
            raise ValueError(f"Group key '{group_key}' not found in pattern.")
        return self.findall(string)[group_key]

    @property
    def example_match(self) -> "MatchResult":
        """Return the match result of the example string and the pattern."""
        return self.match(string=self.example_str)

    @validator("group_keys")
    def validate_group_keys(cls, group_keys, values) -> List[str]:
        """Validate that the number of group keys matches the number of match groups,
        defined in the pattern."""
        pattern = values.get("pattern")
        group_count = count_match_groups(pattern)
        # Subtract non-capturing group count
        if len(group_keys) != group_count:
            raise ValueError(
                f"The number of group keys ({len(group_keys)}) does not match "
                f"the number of match groups ({group_count})."
            )
        return group_keys

    def test_pattern(self) -> bool:
        """Test the pattern by asserting self.example_match.groups.values()) ==
        self.expected_groups"""
        if list(self.example_match.groups.values()) == self.expected_groups:
            return True
        else:
            return False


class MatchResult(OswBaseModel):
    match: Union[re.Match, None]
    pattern: Union[RegExPatternExtended, None]

    class Config:
        arbitrary_types_allowed = True

    @property
    def groups(self):
        """Return a dictionary representation of the object, enabling accessing the
        match groups by key (as defined in RegExPatternExtended)."""
        if self.match is None or self.pattern is None:
            return None  # []
        keys = self.pattern.group_keys
        return {
            key: get_match_group_by_key(match_result=self, group_key=key)
            for key in keys
        }


class SearchResult(MatchResult):
    pass


# Functions
def count_match_groups(pattern: Union[str, re.Pattern]):
    """Counts the number of match groups in a regular expression pattern.

    Parameters
    ----------
    pattern:
        The regular expression pattern.

    Returns
    -------
    group_count:
        The number of match groups in the pattern.

    Examples
    --------
    >>> pat = r"(\\d+)(\\w+)\\((\\d+)\\)"
    >>> grp_cnt = count_match_groups(pattern)
    >>> print(f"Number of match groups: {group_count}")
    """
    if isinstance(pattern, re.Pattern):
        pattern = pattern.pattern
    # Remove escaped parentheses from the pattern
    pattern = re.sub(r"\\[()]", "", pattern)

    # Find all opening and closing parentheses
    open_brackets = re.findall(r"\(", pattern)
    close_brackets = re.findall(r"\)", pattern)

    # Count the number of valid match groups
    group_count = 0
    unmatched_brackets = 0

    for _ in open_brackets:
        unmatched_brackets += 1

    for _ in close_brackets:
        if unmatched_brackets > 0:
            group_count += 1
            unmatched_brackets -= 1

    corrected_group_count = group_count - pattern.count("(?:")
    return corrected_group_count


def match_regex_patterns(
    patterns: Union[RegExPatternExtended, List[RegExPatternExtended]],
    strings: Union[str, List[str]],
    stop_at_first_match: bool = False,
) -> Dict[str, List[MatchResult]]:
    """Match a list of strings against a list of regular expression patterns. If
    stop_at_first_match is True, the function will stop matching a string as soon as it
    finds a match."""
    if isinstance(patterns, RegExPatternExtended):
        patterns = [patterns]
    if isinstance(strings, str):
        strings = [strings]
    matching_results = dict()
    for _k, string in enumerate(strings):
        string_matched = False
        matching_results[string] = list()
        for _i, pattern in enumerate(patterns):
            match = re.match(pattern.pattern, string)
            if match is not None:
                string_matched = True
                matching_results[string].append(
                    MatchResult(match=match, pattern=pattern)
                )
                if stop_at_first_match is True:
                    break
        if string_matched is False:
            matching_results[string].append(MatchResult(match=None, pattern=None))
    return matching_results


def match_first_regex_pattern(
    patterns: Union[RegExPatternExtended, List[RegExPatternExtended]],
    strings: Union[str, List[str]],
) -> List[MatchResult]:
    """Match a list of strings against a list of regular expression patterns. Return
    the first match for each string."""
    if isinstance(patterns, RegExPatternExtended):
        patterns = [patterns]
    if isinstance(strings, str):
        strings = [strings]
    matching_results = match_regex_patterns(patterns, strings, stop_at_first_match=True)
    return [matching_results[string][0] for string in strings]


def get_match_group_by_key(
    match_result: MatchResult, group_key: str
) -> Union[str, None]:
    """Get the value of a named group in a match result."""
    if match_result.match is None:
        return None
    if group_key not in match_result.pattern.group_keys:
        raise ValueError(f"Group '{group_key}' not found in pattern.")
    # +1 is very important here, because the first group is the whole match
    group_index = match_result.pattern.group_keys.index(group_key) + 1
    return match_result.match.group(group_index)


def test_regex_pattern(
    patterns: Union[
        str, List[str], dict, RegExPatternExtended, List[RegExPatternExtended]
    ],
    strings: List[str],
) -> dict:
    """Tests one or a list of regular expression patterns against a list of strings.
    Returns a dictionary with the results. The first pattern is evaluated first,
    and every following pattern is evaluated regarding the additional matches found.

    Parameters
    ----------
    patterns:
        A regular expression pattern string, a list of such strings or a dictionary
        as in osw.data_import_utility.REGEX_PATTERNS
    strings:
        A list of strings to test the patterns against.

    Returns
    -------

    """
    if isinstance(patterns, str):
        patterns = [
            RegExPatternExtended(
                pattern=patterns,
                description="No description",
                group_keys=[],
            )
        ]
    if isinstance(patterns, list):
        if isinstance(patterns[0], str):
            patterns = [
                RegExPatternExtended(
                    pattern=pattern,
                    description="No description",
                    group_keys=[],
                )
                for pattern in patterns
            ]
    if isinstance(patterns, dict):
        patterns = [
            RegExPatternExtended(
                pattern=val["Pattern"],
                description=key,
                group_keys=val["Groups"],
                example_str=val.get("Example string", None),
            )
            for key, val in patterns.items()
        ]
    matching_results = list()
    for _i, pattern in enumerate(patterns):
        result = {
            "Pattern": pattern.pattern,
            "Matching strings": [],
            "Match objects": [],
            "No match": [],
            "Pattern description": pattern.description,
            "Group keys": pattern.group_keys,
        }
        for string in strings:
            match = re.match(pattern.pattern, string)
            match_result = MatchResult(match=match, pattern=pattern)
            if match is None:
                result["No match"].append(string)
            else:
                result["Matching strings"].append(string)
                result["Match objects"].append(match_result)
        result["Coverage per pattern"] = len(result["Matching strings"]) / len(strings)
        matching_results.append(result)

    coverage = dict()
    former_matches = list()
    for i, pattern in enumerate(patterns):
        if pattern.description != "No description":
            key = pattern.description
        else:
            key = i
        previously_unmatched = list(
            set(matching_results[i]["Matching strings"]) - set(former_matches)
        )
        coverage[key] = {
            "Pattern": pattern,
            "Pattern description": pattern.description,
            "Group keys": pattern.group_keys,
            "Matching strings": matching_results[i]["Matching strings"],
            "Matches, groups": [
                [group for group in match_result.match.groups()]
                for match_result in matching_results[i]["Match objects"]
            ],
            "Additional matches": previously_unmatched,
            "Additional matches, groups": [
                [group for group in match_result.match.groups()]
                for match_result in matching_results[i]["Match objects"]
                if match_result.match.string in previously_unmatched
            ],
            "Additional coverage": len(previously_unmatched) / len(strings),
            "Coverage per pattern": matching_results[i]["Coverage per pattern"],
            "Cumulative coverage": (len(former_matches) + len(previously_unmatched))
            / len(strings),
        }
        former_matches.extend(previously_unmatched)

    coverage["Accumulative coverage"] = sum(
        [len(result["Matching strings"]) for result in matching_results]
    ) / len(strings)
    total_coverage = len(former_matches) / len(strings)
    coverage["Total coverage"] = total_coverage
    coverage["Not covered"] = 1 - total_coverage
    coverage["Unmatched strings"] = list(set(strings) - set(former_matches))

    return {"Results of regex matching": matching_results, "Coverage": coverage}



import copy

import mwparserfromhell
import numpy as np
from jsonpath_ng.ext import parse

"""
Module to transform wikitext / templates.
The extra dependency 'osw[wikitext]' is required for this module.
"""


def update_template_within_wikitext(
    text,
    template_text,
    delete=False,
    remove_empty_lines=False,
    overwrite_with_empty=False,
):
    """Updates the template parameters in an existing wiki <text> with a provided new
    <template_text>

    Parameters
    ----------
    text : str
        The existing text
    template_text : str
        The template text with new data
    delete : bool
        If true, params not defined in <template_text> get removed from <text>
    remove_empty_lines : bool
        If true, function will clean-up empty lines within the template code created by
        the underlying mwparserfromhell lib (wanted), but also within the wiki text
        around it (unwanted)
    overwrite_with_empty : bool
        If true, parameters in the existing template will be overwritten even if the
        parameter value in the template_text is empty

    Returns
    -------
    new_text : str
    """
    if template_text == text:
        return template_text
    if template_text == "":
        return text
    if text == "":
        return template_text
    new_code = mwparserfromhell.parse(template_text)
    new_template = new_code.filter_templates()[0]
    existing_code = mwparserfromhell.parse(text)
    template_names_matched = False
    for template in existing_code.filter_templates(recursive=True):
        if template.name.matches(new_template.name):
            existing_template = template
            template_names_matched = True
            break
    if template_names_matched:
        for p in new_template.params:
            if existing_template.has(p.name):
                if p.value != "" or overwrite_with_empty:
                    existing_template.get(p.name).value = p.value
            else:
                existing_template.add(p.name, p.value)
        if delete:
            for p in existing_template.params:
                if not new_template.has(p.name):
                    existing_template.remove(p)
    else:  # the original text did not contain a matching template
        # options:
        # 1) include the new template
        # 2) replace the existing template - but what if multiple tempaltes exist on
        #    that page?
        pass
    # print("Tmerged: \n" + str(existing_template))
    # print("Text merged: " + str(existing_code))
    new_text = str(existing_code)
    # this will cleanup empty lines within the template code (wanted), but also within
    # the wiki text around it (unwanted)
    if remove_empty_lines:
        new_text = "\n".join(
            [ll.rstrip() for ll in str(new_text).splitlines() if ll.strip()]
        )
    return new_text


def merge_wiki_page_text(
    text1, text2, template_name, subtemplate_param="", subtemplate_name=""
):
    """Not fully tested function!

    Parameters
    ----------
    text1
    text2
    template_name
    subtemplate_param
    subtemplate_name

    Returns
    -------
    non_empty_lines : str
    """

    if text1 == text2:
        return text1
    if text1 == "":
        return text2
    if text2 == "":
        return text1
    code1 = mwparserfromhell.parse(text1)
    code2 = mwparserfromhell.parse(text2)
    r1 = []
    r2 = []
    for template in code1.filter_templates(recursive=True):
        if template.name.matches(template_name):
            t1 = template
        if template.name.matches(subtemplate_name):
            r1.append(template)
    for template in code2.filter_templates(recursive=True):
        if template.name.matches(template_name):
            t2 = template
        if template.name.matches(subtemplate_name):
            r2.append(template)
    # print("T1: " + str(t1))
    # print("T2: " + str(t2))
    for p in t2.params:
        if not t1.has(p.name):
            t1.add(p.name, p.value)
    if subtemplate_param != "":
        t1.remove(subtemplate_param)
        for rel2 in r2:
            exists = False
            for rel1 in r1:
                all_params_equal = True
                for p in rel2.params:
                    if rel1.has(p.name):
                        if not rel1.get(p.name).value.matches(rel2.get(p.name).value):
                            all_params_equal = False
                        # print("Compare {}:{} with :{} -> res={}".format(p.name,
                        #       rel1.get(p.name).value,rel2.get(p.name).value,
                        #       rel1.get(p.name).value.matches(rel2.get(p.name).value)))
                    else:
                        all_params_equal = False
                if all_params_equal:
                    exists = True
            if not exists:
                r1.append(rel2)
        r1string = ""
        for rel1 in r1:
            r1string += "\r\n   " + str(rel1)
        t1.add(subtemplate_param, r1string)
    # print("Tmerged: " + str(t1))
    non_empty_lines = "\n".join(
        [ll.rstrip() for ll in str(t1).splitlines() if ll.strip()]
    )
    return non_empty_lines


def create_flat_content_structure_from_wikitext(
    text: str, array_mode: str = "force"
) -> list:
    """Create a flat python dict (aka 'flat_content_structure' = 'wikiJson')
    representing the content of the page

    Parameters
    ----------
    text :
        the wiki source text
    array_mode :
        defines how to parse template params
        array_mode / param     value   value;  value;value     comment
        'force':               array   array   array           always create an array
        'only_multiple':       literal literal array           create only when more
                                                               than one value is given
        'delimiter_present':   literal array   array           create array if at least
                                                               one separator char is
                                                               present

    Returns
    -------
    res :
        (aka 'flat_content_structure' = 'wikiJson')
    """

    res = []
    text = text.strip()
    existing_code = mwparserfromhell.parse(text)
    t_count = 0
    for _i in range(0, len(existing_code.nodes)):
        n = existing_code.nodes.pop(
            0
        )  # returns first layer of nodes. filter() returns also template args
        if type(n) is mwparserfromhell.nodes.template.Template:
            t_count += 1
            wt = {}
            # print(f"Template: {t.name} = {t}")
            wt[str(n.name).strip()] = {}
            for p in n.params:
                # print(f"  Param: {p.name} = {p.value} ({type(p.value)})")
                wt[str(n.name).strip()][str(p.name)] = (
                    create_flat_content_structure_from_wikitext(
                        str(p.value), array_mode
                    )
                )
            res.append(wt)
        else:
            if len(res) == 0 or type(res[-1]) is dict:
                res.append("")
            res[-1] = res[-1] + str(n)  # append to previous string if exists
    # for i, x in enumerate(res):
    #    if type(x) is str: res[i] = x.strip() #remove whitespace
    res = [x for x in res if x and not (type(x) is str and x.isspace())]
    if t_count == 0:
        res = text
        values = str(text).strip().split(";")
        if array_mode == "force":
            res = values
        elif array_mode == "only_multiple" and len(values) > 1:
            res = values
        elif array_mode == "separator_present" and ";" in text:
            res = values
    return res


def find_dependencies(wikitext, debug=False):
    """Finds templates, properties and categories within wikitext

    Parameters
    ----------
    wikitext : str
    debug : bool
        Whether to print debugging messages

    Returns
    -------
    filtered_dependencies : list
    """
    dependencies = []
    code = mwparserfromhell.parse(wikitext)
    for template in code.filter_templates(recursive=True):
        if template.name.split(":")[0].isupper():
            if debug:
                print("MagicWord: {}".format(template.name))
        elif template.name[0] == "#":
            if debug:
                print("ParserFunction: {}".format(template.name))
            if "#set:" in template.name or "#declare:" in template.name:
                if (
                    "=" in template.name.split(":")[1]
                ):  # in case of '{{#set:HasIdPostfix={{{id_postfix}}} }}'
                    property_ = "Property:" + template.name.split(":")[1].split("=")[0]
                    dependencies.append(property_)
                    if debug:
                        print("=> {}".format(property_))
                for param in template.params:
                    property_ = "Property:" + param.split("=")[0]
                    dependencies.append(property_)
                    if debug:
                        print("=> {}".format(property_))
        else:
            if debug:
                print("Template: {}".format(template.name))
            template_name = str(template.name)
            if ":" not in template.name:
                template_name = "Template:" + template_name
            dependencies.append(template_name)
            if debug:
                print("=> {}".format(template_name))
    # for tag in code.filter_tags(recursive=True):
    #    if (debug): print("Tag: {}".format(tag))
    for link in code.filter_wikilinks(recursive=True):
        if "::" in link:
            if debug:
                print("Annotation: {}".format(link))
            property_ = "Property:" + link.split("::")[0].split("[[")[-1]
            dependencies.append(property_)
            if debug:
                print("=> {}".format(property_))
        if "Category:" in link:
            if debug:
                print("Category: {}".format(link))
            category = link.replace("[[", "").replace("]]", "")
            dependencies.append(str(category))
            if debug:
                print("=> {}".format(category))
        else:
            if debug:
                print("Link: {}".format(link))
    dependencies = np.unique(dependencies).tolist()  # remove duplicates
    filtered_dependencies = []  # do not manipulate the iterated object
    for dependency in dependencies:
        # ensure no leading or trailing white spaces
        dependency = dependency.strip()
        # very frew page titles listed here contain invalid characters "\n"
        dependency = dependency.replace("\n", "")
        # see https://www.semantic-mediawiki.org/wiki/Help:Special_properties
        if "Property:" in dependency and (" " in dependency or "_" in dependency):
            if debug:
                print(
                    "Info: Remove presumptive built-in property {}".format(dependency)
                )
        else:
            filtered_dependencies.append(dependency)
    return filtered_dependencies


def find_dependencies_recursively(title, site, dependencies=None, debug=False):
    """Finds all included templates, properties and categories within a wiki page

    Parameters
    ----------
    title : str
        Title of the wiki page, e.g., User:Someone1234
    site : mwclient.client.Site
        Site object from mwclient lib
    dependencies
        List of already found dependent titles
    debug : bool
        Whether to print debugging messages

    Returns
    -------
    dependencies : list
    """
    if dependencies is None:
        dependencies = []
    title = title.replace("{", "").replace("}", "")
    page = site.pages[title]
    content = page.text()
    new_dependencies = find_dependencies(content, debug)
    for dependency in new_dependencies:
        if dependency not in dependencies:  # for circular dependencies
            dependencies.append(dependency)
            if debug:
                print("Scan nested {}".format(dependency))
            find_dependencies_recursively(
                dependency, site, dependencies=dependencies, debug=debug
            )
    return dependencies


def extend_page_list_with_recursive_dependencies(page_list, site):
    """Finds all included templates, properties and categories for a list of wiki pages
    and appends them to the list

    Parameters
    ----------
    page_list : list or iterable
        List of pages
    site : mwclient.client.Site
        Site object from mwclient lib

    Returns
    -------
    page_list : list
    """
    page_list_iter = copy.deepcopy(page_list)
    for page in page_list_iter:
        # we provide the updates page list to prevent a revisit of the pages
        dependencies = find_dependencies_recursively(
            page, site, dependencies=page_list, debug=False
        )
        page_list.extend(dependencies)
    page_list = np.unique(page_list)
    return page_list


def get_wikitext_from_flat_content_dict(d: dict) -> str:
    """Create wiki source text from a flat python dict representing a wiki template

    Parameters
    ----------
    d :
        flat python dict
        e.g.: {"HeaderTemplate": {"param": "value"}}

    Returns
    -------
    wt :
        wiki text
    """
    wt = ""
    for key, value in d.items():
        # print("key: {}, valuetype: {}, value: {}".format(key, type(value), ""))
        if isinstance(value, dict):
            # print("dict")
            wt += "{{" + key
            wt += get_wikitext_from_flat_content_dict(value)
            wt += "\n}}"
        elif isinstance(value, list):
            # print("list")
            wt += "\n|{}=".format(key)
            string_index = 0
            for index, element in enumerate(value):
                if isinstance(element, dict):
                    wt += get_wikitext_from_flat_content_dict(element)
                    # wt += "\n{{" + element
                    # wt += get_wikitext_from_dict(element)
                    # wt += "\n}}"
                else:
                    if string_index != index:
                        print(
                            f"Warning: template param '{key}' has mixed template/"
                            f"string values: {value}"
                        )
                    if string_index > 0 and element and not element.strip().isspace():
                        wt += ";"
                    wt += element
                    string_index += 1
        else:
            # print("literal")
            wt += "\n|{}={}".format(key, value)
    return wt


def get_wikitext_from_flat_content_structure(content):
    """Create wiki source text from the content list (aka 'flat_content_structure' =
    'wikiJson') of the page

    Parameters
    ----------
    content : list
        content list, mixed objects (templates) and free text
        (aka 'flat_content_structure' = 'wikiJson')
        e.g.: [{"HeaderTemplate": {"param": "value"}}, "freetext",
               {"FooterTemplate": {"param2": "value"}}]

    Returns
    -------
    wt : wiki text
    """
    wt = ""
    for content_element in content:
        if isinstance(content_element, dict):
            wt += get_wikitext_from_flat_content_dict(content_element)
        elif isinstance(content_element, str):
            wt += content_element  # "\n" + content_element
        else:
            print(
                "Error: content element is not dict or string: {}".format(
                    content_element
                )
            )
    return wt


def wikiJson2SchemaJson(schema, wikiJson):
    """Create osl schema-compatible json from the content representation of a page
    (aka 'flat_content_structure' = 'wikiJson')

    Parameters
    ----------
    wikiJson : list
        content list, mixed objects (templates) and free text
        (aka 'flat_content_structure' = 'wikiJson')
        e.g.: [{"HeaderTemplate": {"param": "value"}}, "freetext",
               {"FooterTemplate": {"param2": "value"}}]

    Returns
    -------
    schemaJson : dict
        schema-compatible json
        e.g.: {"osl_template": "HeaderTemplate", "param": "value",
               "osl_wikitext": "freetext",
               "osl_footer": {"osl_template": "FooterTemplate",
                              "param2": "value2"}}
    """
    schemaJson = {}
    if (
        not isinstance(wikiJson[0], dict)
        or not isinstance(wikiJson[1], str)
        or not isinstance(wikiJson[2], dict)
    ):
        print("Error: Invalid wikiJson:", wikiJson)
        return schemaJson

    schemaJson = wikiJson2SchemaJsonRecursion(schema, wikiJson[0], wikiJson[2])
    schemaJson["osl_wikitext"] = wikiJson[1]
    return schemaJson


def wikiJson2SchemaJsonRecursion(schema: dict, wikiJson: dict, footerWikiJson=None):
    """internal recursion function of wikiJson2SchemaJson()

    Parameters
    ----------
    wikiJson : list
        wiki template representation
        e.g.: {"HeaderTemplate": {"param": "value"}}

    footerWikiJson : list
        wiki footer template representation
        e.g.: {"FooterTemplate": {"param2": "value2"}}

    Returns
    -------
    schemaJson : dict
        schema-compatible json
        e.g.: {"osl_template": "HeaderTemplate", "param": "value",
               "osl_footer": {"osl_template": "FooterTemplate", "param2": "value2"}}
    """
    schemaJson = {}
    if footerWikiJson is not None:
        schemaJson["osl_footer"] = wikiJson2SchemaJsonRecursion(schema, footerWikiJson)

    for key in wikiJson:
        value = wikiJson[key]
        if isinstance(value, list):
            schemaJson[key] = []
            for index, element in enumerate(value):
                element = value[index]
                if isinstance(element, dict):
                    if key == "extensions":
                        if (
                            footerWikiJson is not None
                        ):  # we asume that every extension provides also a footer template
                            nextFooter = footerWikiJson[
                                schemaJson["osl_footer"]["osl_template"]
                            ]["extensions"][index]
                            schemaJson[key].append(
                                wikiJson2SchemaJsonRecursion(
                                    schema, element, nextFooter
                                )
                            )
                    else:
                        schemaJson[key].append(
                            wikiJson2SchemaJsonRecursion(schema, element)
                        )
                else:
                    schemaJson[key].append(element)

        elif isinstance(value, dict):
            schemaJson = wikiJson2SchemaJsonRecursion(schema, value, footerWikiJson)
            schemaJson["osl_template"] = key
        else:
            schemaJson[key] = value

    if "osl_template" in schemaJson:
        jsonpath_expr = parse(
            "$..properties"
        )  # [?osl_template.default = "' + schemaJson['osl_template'] + '"]')
        schema_def = {}
        # print("Search for ", schemaJson['osl_template'])
        for match in jsonpath_expr.find(schema):
            value = match.value
            if (
                "osl_template" in value
                and "osl_template" in schemaJson
                and value["osl_template"]["default"] == schemaJson["osl_template"]
                or "osl_schema" in value
                and "osl_schema" in schemaJson
                and value["osl_schema"]["default"] == schemaJson["osl_schema"]
            ):
                schema_def = match.value  # ToDo: Resolve allOf...
                # print(schema_def)

        for key in list(schemaJson.keys()):
            if key in schema_def:
                if "type" in schema_def[key]:  # use schema to resolve types
                    if schema_def[key]["type"] == "integer":
                        schemaJson[key] = int(schemaJson[key])
                    elif schema_def[key]["type"] == "float":
                        schemaJson[key] = float(schemaJson[key])
                    elif schema_def[key]["type"] == "number":
                        schemaJson[key] = float(schemaJson[key])
                    elif schema_def[key]["type"] == "string":
                        schemaJson[key] = str(schemaJson[key])
                    elif schema_def[key]["type"] == "array":
                        if not isinstance(schemaJson[key], list):
                            del schemaJson[key]
                        elif len(schemaJson[key]) == 0:
                            del schemaJson[key]
                else:  # assume type==object
                    if isinstance(schemaJson[key], list) and len(schemaJson[key]) == 1:
                        schemaJson[key] = schemaJson[key][0]
            else:  # fall back
                if schemaJson[key] == "" and key == "extensions":
                    del schemaJson[key]
                elif isinstance(
                    schemaJson[key], list
                ):  # wikiJson defaults are lists, even for single or empty values
                    if len(schemaJson[key]) == 0:
                        del schemaJson[key]
                    # elif len(schemaJson[key]) == 1: schemaJson[key] = schemaJson[key][0]

    return schemaJson


def schemaJson2WikiJson(schemaJson, isRoot=True):
    """Create content representation of a page
    (aka 'flat_content_structure' = 'wikiJson') from the osl schema-compatible json

    Parameters
    ----------
    schemaJson : dict
        schema-compatible json
        e.g.: {"osl_template": "HeaderTemplate", "param": "value",
               "osl_wikitext": "freetext",
               "osl_footer": {"osl_template": "FooterTemplate", "param2": "value2"}}

    isRoot: boolean
        indicates first call in recursion

    Returns
    -------
    wikiJson : list
        content list, mixed objects (templates) and free text
        (aka 'flat_content_structure' = 'wikiJson')
        e.g.: [{"HeaderTemplate": {"param": "value"}}, "freetext",
               {"FooterTemplate": {"param2": "value"}}]
    """
    wikiJson = [{}, "", {}]
    # header, freetext, footer
    template = ""
    footer_template = ""
    if "osl_template" in schemaJson:
        template = schemaJson["osl_template"]
        wikiJson[0][template] = {}
    else:
        print(
            "Error: Mandatory property 'osl_template' not found in schemaJson",
            schemaJson,
        )
        return

    if "osl_wikitext" in schemaJson:
        wikiJson[1] = schemaJson["osl_wikitext"]
    if "osl_footer" in schemaJson:
        wikiJson[2] = schemaJson2WikiJson(schemaJson["osl_footer"], False)[0]
        footer_template = schemaJson["osl_footer"]["osl_template"]
        wikiJson[2][footer_template]["extensions"] = []

    for key in schemaJson:
        if (
            key.startswith("_")
            or key.startswith("osl_template")
            or key.startswith("osl_wikitext")
            or key.startswith("osl_footer")
        ):
            continue  # exclude private and reserved keywords
        if schemaJson[key] is None:
            continue
        if isinstance(schemaJson[key], list):
            wikiJson[0][template][key] = []
            for subSchemaJson in schemaJson[key]:
                subWikiJson = schemaJson2WikiJson(subSchemaJson, False)
                wikiJson[0][template][key].append(subWikiJson[0])
                if key == "extensions":
                    wikiJson[2][footer_template]["extensions"].append(subWikiJson[2])

        elif isinstance(schemaJson[key], dict):
            subWikiJson = schemaJson2WikiJson(schemaJson[key], False)
            wikiJson[0][template][key] = [subWikiJson[0]]
        else:
            wikiJson[0][template][key] = schemaJson[key]

    return wikiJson



import json

from jsondiff import diff

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.core import OSW
from osw.utils.wiki import get_full_title
from osw.wtsite import WtSite


# run with: tox -e test -- --wiki_domain domain --wiki_username user --wiki_password pass
def test_local_rdf_graph_with_remote_context(wiki_domain, wiki_username, wiki_password):
    cm = CredentialManager()
    cm.add_credential(
        CredentialManager.UserPwdCredential(
            iri=wiki_domain, username=wiki_username, password=wiki_password
        )
    )
    wtsite = WtSite(WtSite.WtSiteConfig(iri=wiki_domain, cred_mngr=cm))
    osw_obj = OSW(site=wtsite)

    i1 = model.Item(name="MyItem", label=[model.Label(text="My Item")])
    i2 = model.Item(
        name="MyItem2",
        label=[model.Label(text="My Item 2")],
        based_on=[get_full_title(i1)],
    )

    res = osw_obj.export_jsonld(
        params=OSW.ExportJsonLdParams(
            entities=[i1, i2],
            mode=OSW.JsonLdMode.expand,
            build_rdf_graph=True,
            additional_context="https://json-ld.org/contexts/remote-context.jsonld",
        )
    )

    qres = res.graph.query(
        """
        SELECT ?pname ?oname
        WHERE {
        ?s <https://www.w3.org/TR/skos-reference/isBasedOn> ?o .
        ?s Property:HasName ?pname .
        ?o Property:HasName ?oname
        }
        """
    )

    # export Sparql query results as json
    actual = json.loads(qres.serialize(format="json"))
    expected = {
        "results": {
            "bindings": [
                {
                    "oname": {"type": "literal", "value": "MyItem"},
                    "pname": {"type": "literal", "value": "MyItem2"},
                }
            ]
        },
        "head": {"vars": ["pname", "oname"]},
    }
    diff_dict = diff(expected, actual, syntax="explicit")
    print("expected - actual diff: ", diff_dict)
    assert actual == expected



import json
from typing import Any, Dict, Union
from uuid import UUID

import osw.model.entity as model
from osw.core import OSW, AddOverwriteClassOptions, OverwriteOptions
from osw.utils.wiki import remove_empty
from osw.wtsite import WtPage


class OfflineWtPage(WtPage):
    def __init__(self, wtSite: Any = None, title: str = None):
        self.wtSite = wtSite
        self.title = title

        self.exists = True  # only to fake the existence of the page for testing
        self._original_content = ""
        self._content = ""
        self.changed: bool = False
        self._dict = []
        self._slots: Dict[str, Union[str, dict]] = {"main": ""}
        self._slots_changed: Dict[str, bool] = {"main": False}
        self._content_model: Dict[str, str] = {"main": "wikitext"}


def test_osw_id_to_uuid():
    osw_id = "OSW2ea5b605c91f4e5a95593dff79fdd4a5"
    expected = UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")
    result = OSW.get_uuid(osw_id)
    assert result == expected


def test_uuid_to_osw_id():
    uuid = UUID("2ea5b605-c91f-4e5a-9559-3dff79fdd4a5")
    expected = "OSW2ea5b605c91f4e5a95593dff79fdd4a5"
    result = OSW.get_osw_id(uuid)
    assert result == expected


def check_true(original: model.Entity, altered: model.Entity, stored: model.Entity):
    """Check the entity that was stored with the 'overwrite' param set to True,
    which is supposed to overwrite all existing properties of the entity in the
    OSW, but keep those not present in the altered entity."""
    assert stored.label[0].text == altered.label[0].text
    assert stored.name == altered.name
    assert stored.iri == altered.iri
    assert stored.description[0].text == altered.description[0].text
    assert stored.query_label == altered.query_label
    assert stored.image == original.image
    assert stored.attachments == altered.attachments


def check_false(original: model.Entity, altered: model.Entity, stored: model.Entity):
    """Check the entity that was stored with the 'overwrite' param set to False,
    which is supposed to keep all existing properties of the entity in the OSW,
    but add those additionally present in the altered entity."""
    assert stored.label[0].text == original.label[0].text
    assert stored.name == original.name
    # empty string property is removed on store and load:
    assert stored.iri == altered.iri
    if original.description is None or len(original.description) == 0:
        assert stored.description == altered.description
    else:
        assert stored.description[0].text == original.description[0].text
    assert stored.query_label == altered.query_label  # value == None -->
    # property not present and will be set (overwritten)
    assert stored.image == original.image
    assert stored.attachments == altered.attachments


def check_only_empty(
    original: model.Entity, altered: model.Entity, stored: model.Entity
):
    """Check the entity that was stored with the 'overwrite' param set to
    'only empty', which is supposed to overwrite only those properties of the
    entity in the OSW that are empty, but are not empty in the altered entity."""
    assert stored.label[0].text == original.label[0].text
    assert stored.name == original.name
    # empty string property is removed on store and load:
    assert stored.iri == altered.iri
    assert stored.description[0].text == altered.description[0].text
    assert stored.query_label == altered.query_label
    assert stored.image == original.image
    assert stored.attachments == altered.attachments


def check_replace_remote(
    original: model.Entity, altered: model.Entity, stored: model.Entity
):
    """Check the entity that was stored with the 'overwrite' param set to
    'replace remote', which is supposed to replace the remote entity entirely."""
    assert stored.label[0].text == altered.label[0].text
    assert stored.name == altered.name
    assert stored.iri == altered.iri
    assert stored.description[0].text == altered.description[0].text
    assert stored.query_label == altered.query_label
    assert getattr(stored, "image", None) is None
    assert stored.attachments == altered.attachments


def check_keep_existing(
    original: model.Entity, altered: model.Entity, stored: model.Entity
):
    """Check the entity that was stored with the 'overwrite' param set to
    'keep existing', which is supposed to keep the existing entity in the OSW."""
    assert stored.label[0].text == original.label[0].text
    assert stored.name == original.name
    # assert stored.iri == original.iri
    # empty string property is removed on store and load:
    assert getattr(stored, "iri", None) is None
    if original.description is None or len(original.description) == 0:  # empty list
        assert stored.description is None
    else:
        assert stored.description[0].text == original.description[0].text
    assert stored.query_label == original.query_label
    assert stored.image == original.image
    assert getattr(stored, "attachments", None) is None


def test_apply_overwrite_policy():
    checks = [
        {  # Overwrite properties
            "overwrite": OverwriteOptions.true,
            "assert": check_true,
        },
        {  # Do not overwrite properties
            "overwrite": OverwriteOptions.false,
            "assert": check_false,
        },
        {  # Overwrite empty properties only
            "overwrite": OverwriteOptions.only_empty,
            "assert": check_only_empty,
        },
        {  # Replace the remote entity entirely
            "overwrite": AddOverwriteClassOptions.replace_remote,
            "assert": check_replace_remote,
        },
        {  # Keep the existing entity as is
            "overwrite": AddOverwriteClassOptions.keep_existing,
            "assert": check_keep_existing,
        },
    ]

    for check in checks:
        # Create a new item with some properties
        original_item_local = model.Item(
            label=[model.Label(text="My Item")],
            name="MyItem",
            iri="",  # Empty string property
            description=[],  # Empty list property
            query_label=None,  # Equal to a non-existing property
            image="File:OSWacacwdcawd.png",  # Property not in altered
        )
        original_page = OfflineWtPage(
            # wtSite=OSW.wt_site,  # todo: missing
            title="Item:"
            + OSW.get_osw_id(original_item_local.uuid)
        )
        jsondata = json.loads(original_item_local.json(exclude_none=True))
        # Emulate the default setting for StoreEntityParam.remove_empty,
        #  which is True and applied on store_entity
        remove_empty(jsondata)

        original_page.set_slot_content("jsondata", jsondata)
        # To reproduce what happens in the OSL instance, we need to get the content
        #  from the page again
        original_content = original_page.get_slot_content("jsondata")
        # And create the original item from that content
        original_item = model.Item(**original_content)
        # Alter some of the property values
        altered_props = {
            "label": [model.Label(text="My Item Duplicate")],
            "name": "MyItemDuplicate",
            "iri": "http://example.com/MyItemDuplicate",
            "description": [model.Label(text="This is a duplicate")],
            "query_label": "My Item Duplicate",
            "attachments": ["File:OSWacacwdcawd.pdf"],  # Property not in original
        }
        # Create a new item with the altered properties
        original_props = {
            key: value
            for key, value in original_item.dict().items()
            if key in ["uuid", "type"]
        }
        altered_item = model.Item(**{**original_props, **altered_props})
        new_page = OSW._apply_overwrite_policy(
            param=OSW._ApplyOverwriteParam(
                page=original_page,
                entity=altered_item,
                policy=check["overwrite"],
                inplace=False,
            )
        )
        # Check the altered page
        new_content = new_page.get_slot_content("jsondata")
        new_item = model.Item(**new_content)

        check["assert"](original_item, altered_item, new_item)

        # todo: # where do we really need to interact with an OSL instance?


def test_wtpage_get_files():
    # Create an offline page with test content in slot main
    # Real-world example from Item:OSW8bccb1f0123f47d1831a1348ecbe63cc
    # (About this platform)
    page = OfflineWtPage()
    page.set_slot_content(
        "main",
        """

====File pages====
The following image gallery displays the relevant elements of a WikiFile page, which
 are specific to file pages.
{{Template:Viewer/Media
| image_size = 300
| mode = default
| textdata = File:OSW486215598a1f4993b063804775d70716.png{{!}}Example WikiFile page
with preview of the file;
File:OSWb4f81db0862d4430b45e6fcbda9fc1ff.PNG{{!}}Infobox of a WikiFile page;
File:OSWb6115f4d5b414a3f8b3dffd420c82c2e.PNG{{!}}Footer of a WikiFile page;
}}

===Semantic triples===
The semantic triple is the basic building block of semantic technology. It links two
nodes ("Subject" and "Object") by a "Property" (sometimes called "Predicate"), which
expresses the relation between the two, giving meaning to the link.
{{Template:Editor/DrawIO
| file_name = Semantic triple schematic
| page_name = Item:OSW8bccb1f0123f47d1831a1348ecbe63cc
| uuid = c4171917-ea09-4d98-823a-6af8282a6d50
| full_width = 0
| width = 300px
}}{{Template:Viewer/Media
| image_size = 600
| mode = default
| textdata = File:OSW51ad8f9d660641f9880006c40f41cb56.png{{!}}An example for a network
of semantic links, describing a publication, one of the authors and his affiliate;
}}

===Ontologies===
Ontologies aren't just vocabularies that define terms linke a dictionary would do.
Ontologies structure knowledge by defining concepts and the relations among them. Often,
they involve a hierarchy, which springs from a very generic object, aiming to describe
almost everything, like "Thing" in
{{Template:Viewer/Link|page=|url=https://schema.org/docs/full.html|label=Schema.org}}
or "Entity" within this [[:Category:Entity#Subcategories|platform]].
Use the > Symbol to expand different hierarchy levels and to explore the structure.

There are many ontologies in use and even more in development. Most scientific domains
have their own, often multiples. Here are some prominent examples:

*{{Template:Viewer/Link|page=|url=https://emmo-repo.github.io/|label=The Elementary
Multiperspective Material Ontology (EMMO)}}
* {{Template:Viewer/Link|page=|url=https://big-map.github.io/BattINFO/index.html|label=
Battery Interface Ontology (BattINFO)}}

==Object Orientation==
Object Orientation (OO) is a theoretical concept in computer science. It uses the
abstract concept of objects to describe and model real-world objects.

===Basic building blocks===
*Object


{{Template:Editor/DrawIO
| file_name = Basic building blocks of Object Orientation diagramm
| page_name = Item:OSW8bccb1f0123f47d1831a1348ecbe63cc
| uuid = 0bea84d5-4c07-4374-a4b4-5dc84d9ba302
| full_width = 0
| width = 200px
}}

===Linked Data===
Object oriented linked data (OO-LD) in Open Semantic Lab is our way to leverage the
functionality of linked data while employing concepts of object orientation to avoid
the redundant definition of (semantic) properties of objects and mapping of semantic
properties to ontologies.

The following figure sketches how Open Semantic Lab is used to modell objects in the
real-world. On the left we see a hierarchy of abstract concepts, starting from the most
generic at the top and ending up at the most specific at the bottom. At each level
either new properties are introduced or the range of property values is reduced to a
certain set to account for the specifics of a category. A category or class is used to
define and bundle all properties or attributes that items or instances of a certain
class have in common.  {{Template:Editor/DrawIO
| file_name = Object oriented linked data in OSL diagramm
| page_name = Item:OSW8bccb1f0123f47d1831a1348ecbe63cc
| uuid = 58baa09e-c00b-42cc-b077-9fe4d58ccf82
| width = 600px
}}

==Glossary of frequently used terms==
""",
    )

    file_list = page.find_file_page_refs_in_slots()
    print(file_list)
    assert len(file_list) == 7
    # assert if the file list contains the expected files
    assert "File:OSW486215598a1f4993b063804775d70716.png" in file_list
    assert "File:OSWb4f81db0862d4430b45e6fcbda9fc1ff.PNG" in file_list
    assert "File:OSWb6115f4d5b414a3f8b3dffd420c82c2e.PNG" in file_list
    assert "File:OSWc4171917ea094d98823a6af8282a6d50.drawio.svg" in file_list
    assert "File:OSW51ad8f9d660641f9880006c40f41cb56.png" in file_list
    assert "File:OSW0bea84d54c074374a4b45dc84d9ba302.drawio.svg" in file_list
    assert "File:OSW58baa09ec00b42ccb0779fe4d58ccf82.drawio.svg" in file_list



import os
from pprint import pprint
from typing import ClassVar
from uuid import uuid4

from pydantic.v1 import BaseModel

import osw.model.entity as model
from osw.core import OSW
from osw.express import OswExpress

# Create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
osw_obj = OswExpress(
    domain="wiki-dev.open-semantic-lab.org", cred_filepath=pwd_file_path
)


# todo: does dataclass export only most specific class jsonschema
class MyPythonClass(
    BaseModel
):  # We don't inherit from model.Item here because this will trigger a full schema
    #  export
    __uuid__: ClassVar[uuid4] = "23e4356e-b726-4c5b-b63f-620b301eb836"
    my_value: str


osw_obj.register_schema(
    osw_obj.SchemaRegistration(
        model_cls=MyPythonClass,
        schema_uuid=MyPythonClass.__uuid__,
        schema_name="MyPythonClass",
        schema_bases=["Category:Item"],
    )
)

osw_obj.fetch_schema(
    osw_obj.FetchSchemaParam(
        schema_title="Category:" + OSW.get_osw_id(MyPythonClass.__uuid__), mode="append"
    )
)

my_instance = model.MyPythonClass(
    label=[model.Label(text="My Instance")], my_value="test"
)
pprint(my_instance)



from osw.utils.regex import RegExPatternExtended

REGEX_PATTERN_LIST = [
    RegExPatternExtended(
        description="File page full page title (new format) to label",
        pattern=r"File:(OS[WL]{1}[a-f0-9]{32})"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Label", "Suffix(es)"],  # ["OSW-ID", "Suffix(es)"]
        example_str="File:OSW11a53cdfbdc24524bf8ac435cbf65d9d.svg",
        expected_groups=["OSW11a53cdfbdc24524bf8ac435cbf65d9d", ".svg"],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with OSW-ID and two "
        "suffixes",
        pattern=r"File:([A-Za-z]+)[\- ]{1}(OS[WL]{1}[a-f0-9]{32})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*"
        r"\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Namespace", "OSW-ID", "Label", "Suffix(es)"],
        example_str="File:LabNote-OSL9568a776dcdf4b7e86f462993df12059-"
        "spreadsheet-01.luckysheet.json",
        expected_groups=[
            "LabNote",
            "OSL9568a776dcdf4b7e86f462993df12059",
            "spreadsheet-01",
            ".luckysheet.json",
        ],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with OSW-ID and one suffix",
        pattern=r"File:([A-Za-z]+)[\- ]{1}(OS[WL]{1}[a-f0-9]{32})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Namespace", "OSW-ID", "Label", "Suffix(es)"],
        example_str="File:LabNote-OSL9ee686dbcc654f4abeaf5f38b8487885-sketch-05.svg",
        expected_groups=[
            "LabNote",
            "OSL9ee686dbcc654f4abeaf5f38b8487885",
            "sketch-05",
            ".svg",
        ],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with namespace, ISC-ID and "
        "two suffixes",
        pattern=r"File:([A-Za-z]+)[\- ]{1}([A-Za-z0-9\-]+[\- ]{1}[0-9]{4})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*"
        r"\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Namespace", "ISC-ID", "Label", "Suffix(es)"],
        example_str="File:LabNote-220304-test-0002-ni-diagram-01.drawio.svg",
        expected_groups=["LabNote", "220304-test-0002", "ni-diagram-01", ".drawio.svg"],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with namespace, ISC-ID and "
        "one suffix",
        pattern=r"File:([A-Za-z]+)[\- ]{1}([A-Za-z0-9\-]+[\- ]{1}[0-9]{4})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Namespace", "ISC-ID", "Label", "Suffix(es)"],
        example_str="File:LabNote-220305-sist-0001-ni-sketch-furnace-black.svg",
        expected_groups=[
            "LabNote",
            "220305-sist-0001",
            "ni-sketch-furnace-black",
            ".svg",
        ],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with ISC-ID and two "
        "suffixes",
        pattern=r"File:([A-Za-z0-9\-]+[\- ]{1}[0-9]{4})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*"
        r"\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["ISC-ID", "Label", "Suffix(es)"],
        example_str="File:211104-sist-0003-n stamped 20211116080457.pdf.ots",
        expected_groups=["211104-sist-0003", "n stamped 20211116080457", ".pdf.ots"],
    ),
    RegExPatternExtended(
        description="File page full page title (old format) with ISC-ID and "
        "one suffix",
        pattern=r"File:([A-Za-z0-9\-]+[\- ]{1}[0-9]{4})[\- ]{1}(.*)"
        r"(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["ISC-ID", "Label", "Suffix(es)"],
        example_str="File:220304-test-0003-ni stamped 20220304155047.pdf",
        expected_groups=["220304-test-0003", "ni stamped 20220304155047", ".pdf"],
    ),
    RegExPatternExtended(
        description="File page full page title with two suffixes",
        pattern=r"File:(.+)(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*"
        r"\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Label", "Suffix(es)"],
        example_str="File:OpenSemanticLab-vertical-process-abstraction-en.drawio.svg",
        expected_groups=[
            "OpenSemanticLab-vertical-process-abstraction-en",
            ".drawio.svg",
        ],
    ),
    RegExPatternExtended(
        description="File page full page title with one suffix",
        pattern=r"File:(.+)(\.[a-zA-Z0-9]*[a-zA-Z]+[a-zA-Z0-9]*)",
        group_keys=["Label", "Suffix(es)"],
        example_str="File:ISC-Digital-1440x448px.jpg",
        expected_groups=["ISC-Digital-1440x448px", ".jpg"],
    ),
    RegExPatternExtended(
        description="File page strings from any text",
        pattern=r"(File:(OSW[0-9a-fA-F]{32})((?:\.[\w-]+)*))",
        group_keys=["Full page name", "OSW-ID", "Suffix(es)"],
        example_str="File:OSWb701155d41ad44ab9ad719edc3a2480a.drawio.png",
        expected_groups=[
            "File:OSWb701155d41ad44ab9ad719edc3a2480a.drawio.png",
            "OSWb701155d41ad44ab9ad719edc3a2480a",
            ".drawio.png",
        ],
    ),
    RegExPatternExtended(
        description="File uuid in template",
        pattern=r"{{\s*(Template:Editor\/(DrawIO|SvgEdit|Kekule|Spreadsheet))[\s\S]*?\|\s*uuid\s*=\s*([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})[\s\S]*?}}",
        group_keys=["Full template name", "Editor", "UUID"],
        example_str="""{{Template:Editor/DrawIO
            | file_name = Semantic triple schematic
            | page_name = Item:OSW8bccb1f0123f47d1831a1348ecbe63cc
            | uuid = c4171917-ea09-4d98-823a-6af8282a6d50
            | full_width = 0
            | width = 300px
            }}""",
        expected_groups=[
            "Template:Editor/DrawIO",
            "DrawIO",
            "c4171917-ea09-4d98-823a-6af8282a6d50",
        ],
    ),
    RegExPatternExtended(
        description="UUID from full page title",
        pattern=r"([A-Za-z]+):([A-Z]+)([0-9a-fA-F]{32})((?:\.[\w-]+)*)",
        group_keys=["Namespace", "Prefix", "UUID", "Suffix(es)"],
        example_str="File:OSWff61ba91c92742a3ade0ce24c1399cb2.drawio.png",
        expected_groups=[
            "File",
            "OSW",
            "ff61ba91c92742a3ade0ce24c1399cb2",
            ".drawio.png",
        ],
    ),
    RegExPatternExtended(
        description="UUID from OSW ID",
        pattern=r"([A-Z]+)([a-z\d\-]+)",
        group_keys=["Prefix", "UUID"],
        example_str="OSWff61ba91c92742a3ade0ce24c1399cb2",
        expected_groups=["OSW", "ff61ba91c92742a3ade0ce24c1399cb2"],
    ),
]
REGEX_PATTERN = {rep.description: rep.dict() for rep in REGEX_PATTERN_LIST}
REGEX_PATTERN_LIB = {rep.description: rep for rep in REGEX_PATTERN_LIST}



import os

import osw.model.entity as model
from osw.auth import CredentialManager
from osw.express import OswExpress

# Create/update the password file under examples/accounts.pwd.yaml
pwd_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "accounts.pwd.yaml"
)
# pwd_file_path = "./accounts.pwd.yaml"
osw_obj = OswExpress(
    domain="wiki-dev.open-semantic-lab.org", cred_filepath=pwd_file_path
)

# load dependencies
DEPENDENCIES = {
    "Host": "Category:OSW02590972aeba46d7864ed492c0c11384",
    "DatabaseServer": "Category:OSWacdb001c926c46b998af3e645d36b13f",
    "Database": "Category:OSW51ad0d1716254c77a2b7a03217f23b43",
    "DatabaseType": "Category:OSW33c7c3a4076a4a58b20d818c162e84e6",
}
osw_obj.install_dependencies(DEPENDENCIES, mode="append")

# Make sure to import controllers after updating the model (ignore linter warning)
import osw.controller as controller  # noqa: E402

# Load database definition
db = osw_obj.load_entity("Item:OSWb8cc7705e17c47b19331fdb045bfbca8")  # postgres
db = db.cast(model.Database)

# Cast into controller and execute function
db = db.cast(controller.DatabaseController)
db.connect(
    controller.DatabaseController.ConnectionConfig(
        cm=CredentialManager(cred_filepath=pwd_file_path), osw=osw_obj
    )
)
db.execute("SELECT * FROM public.sensors ORDER BY sensor_id ASC")
